#include <stdio.h>

#include "Compiler.h"
#include "Directory/Directory.h"
#include "OS Specific/DirString.h"
#include "Directory/Management.h"
#include "Directory/File.h"

#include "Debug/Debug.h"

Compiler::Compiler(const CString &name)
{
  Name          = name;
  Default       = 0;
  ShowCompFiles = 0;
  NFilesBuilt   = 0;
  Built         = 0;
}


void Compiler::Expand(void)
{
  CompileStack.Expand(CompileStr);
  LinkStack.Expand(LinkStr);
}

void Compiler::Print(void)
{
  printf("      Compiler \"%s\"\n", Name.c_str());
  printf("       Default %d\n",     Default);
  printf(" ShowCompFiles %d\n",     ShowCompFiles);
  printf("       Compile \"%s\"\n", CompileStr.c_str());
  printf("          Link \"%s\"\n", LinkStr.c_str());
  printf("\n");
}

void Compiler::BuildIncludeDirs(Directory *node)
{
  unsigned int i;

  if(!node)
    return;

  VALIDATE_POINTER(node)

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    for(unsigned int i = 0; i < node->Dirs.size(); ++i)
      VALIDATE_POINTER(node->Dirs[i])
  #endif

  if(AutoDirTree && node->Path.length())
    IncludeDirs += IncludeDirSwitch + "\"" + node->Path + "\" ";

  for(i = 0; i < node->Dirs.size(); ++i)
    BuildIncludeDirs(node->Dirs[i]);
}

void Compiler::AddAdditonalIncs(CVector<CString> &add)
{
  unsigned int i;
  for(i = 0; i < add.size(); ++i)
    IncludeDirs += IncludeDirSwitch + "\"" + add[i] + "\" ";
}

void Compiler::Compile(const CString &objdir, Directory *node, ConfigFile *config)
{
  unsigned int i;

  if(!node)
    return;

  VALIDATE_POINTER(node)
  VALIDATE_POINTER(node->Config)

  if(node->Config == config)
  {
    CheckDirRec(objdir + node->Path);

    for(i = 0; i < node->Sources.size(); ++i)
    {
      CString obj = node->Sources[i]->GetObjFile();

      VALIDATE_POINTER(node->Sources[i]);

      if(node->Sources[i]->CompileRequired())
      {
        if(ShowCompFiles) printf("%s\n", node->Sources[i]->Name.c_str());
        remove(obj.c_str());
        CompileStack.Run(Vars(node->Sources[i]->Path + node->Sources[i]->Name, obj, IncludeDirs, "", ""));
        if(!FileExists(obj))
          Errors.push_back(node->Sources[i]->Path + node->Sources[i]->Name);
        else
        {
          ++NFilesBuilt;
          ObjsToLink += '\"';
          ObjsToLink += obj;
          ObjsToLink += "\" ";
        }
      }
      else
      {
        ObjsToLink += '\"';
        ObjsToLink += obj;
        ObjsToLink += "\" ";
      }
    }
  }

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    for(unsigned int i = 0; i < node->Dirs.size(); ++i)
      VALIDATE_POINTER(node->Dirs[i])
  #endif

  for(i = 0; i < node->Dirs.size(); ++i)
    Compile(objdir, node->Dirs[i], config);
}

void Compiler::Link(int depends_built)
{
  if(!CanLink())
    return;

  if(!depends_built && !NFilesBuilt && FileExists(OutDir + FileTitle))
  {
    printf("No detected changes. Skipping linker phase.\n");
    return;
  }

  if(!ObjsToLink.size() && !LibsToLink.size())
    return;

  remove((OutDir + FileTitle).c_str());
  LinkStack.Run(Vars("", ObjsToLink, "", LibsToLink, OutDir + FileTitle));
  Built = 1;
}

int Compiler::CanLink(void)
{
  if(Errors.size())
    return 0;
  return 1;
}

