#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <stdlib.h>

#include "FileIO/Parser.h"
#include "Directory/FilePaths.h"
#include "Directory/Management.h"
#include "Directory/DirectoryTree.h"
#include "OS Specific/DirString.h"
#include "OS Specific/DirList.h"
#include "OS Specific/SystemCalls.h"

#include "ConfigFile.h"

#include "Debug/Debug.h"

static char *AddSkipDir       (Parser &p, char *str, ConfigFile *config);
static char *AddSkipFile      (Parser &p, char *str, ConfigFile *config);
static char *AddIncludeDir    (Parser &p, char *str, ConfigFile *config);
static char *AddHeaderType    (Parser &p, char *str, ConfigFile *config);
static char *AddSourceType    (Parser &p, char *str, ConfigFile *config);
static char *AddMinimumRebuild(Parser &p, char *str, ConfigFile *config);
static char *AddIsLibrary     (Parser &p, char *str, ConfigFile *config);
static char *AddObjectDir     (Parser &p, char *str, ConfigFile *config);
static char *AddLibraryDir    (Parser &p, char *str, ConfigFile *config);
static char *AddBinaryDir     (Parser &p, char *str, ConfigFile *config);
static char *AddCompiler      (Parser &p, char *str, ConfigFile *config);
static char *AddDefault       (Parser &p, char *str, ConfigFile *config);
static char *AddFileTitle     (Parser &p, char *str, ConfigFile *config);
static char *AddShowCompFiles (Parser &p, char *str, ConfigFile *config);
static char *AutoIncDir       (Parser &p, char *str, ConfigFile *config);
static char *IncludeDirSwitch (Parser &p, char *str, ConfigFile *config);
static char *AddObjExt        (Parser &p, char *str, ConfigFile *config);
static char *AddCompileString (Parser &p, char *str, ConfigFile *config);
static char *AddLinkString    (Parser &p, char *str, ConfigFile *config);
static char *LoadDBase        (Parser &p, char *str, Database   *dbase);

ConfigFile::ConfigFile(const CString &file)
{
  F = file;
  Comp = 0;
  if(FileExists(file))
    Read(file.c_str());
  else
    GenerateDefaults(file.c_str());

  ValidateDirs();

  Saved = 0;
}


void ConfigFile::SetCompiler(const CString &comp)
{
  unsigned int i;

  Comp = 0;

  for(i = 0; i < Comps.size(); ++i)
    if(Cmp(Comps[i].Name.c_str(), comp.c_str()))
    {
      Comps[i].Default = 1;
      Comp = &Comps[i];
    }
    else
      Comps[i].Default = 0;
}

int ConfigFile::IsCompiler(const CString &comp)
{
  unsigned int i;

  for(i = 0; i < Comps.size(); ++i)
    if(Comps[i].Name == comp)
      return 1;
  return 0;
}

void ConfigFile::GenerateDefaults(const char *file)
{
  printf("Warning: %s does not exist. Generating Default.\n", file);
  HeaderTypes.push_back("*.h");
  SourceTypes.push_back("*.cpp");
  SourceTypes.push_back("*.c");
  BinDir     = "";
  ObjDir     = "Objs";
  LibDir     = "Libs";
  MinRebuild = 1;
  IsLibrary  = 0;
  Comps.push_back(Compiler("g++"));
  Comps.back().FileTitle        = "a.exe";
  Comps.back().Default          = 1;
  Comps.back().ShowCompFiles    = 1;
  Comps.back().AutoDirTree      = 1;
  Comps.back().IncludeDirSwitch = "-I ";
  Comps.back().ObjExt           = ".o";
  Comps.back().CompileStr       = "g++ -Wall -Wextra -ansi -pedantic -c %INCLUDE DIRS% -o %OBJ% %SOURCE%";
  Comps.back().LinkStr          = "g++ %COMPILED OBJS% %COMPILED LIBS% -lwinmm -o %BIN%";

  Comps.back().Expand();
  Comp = &Comps.back();
  IgnoreDirs.push_back(ObjDir);
  IgnoreDirs.push_back(LibDir);
}

int ConfigFile::CheckDataBase(const CString &file, Directory *root)
{
  unsigned int mod;
    // Pointer for debug related reasons
  File *f = new File(GetPath(file), GetFile(file), root, this);
  if(DBase.FileModified(f))
  {
    delete f;
    return 1;
  }

  f->CheckIncludes(1);
  mod = f->WasModified;
  delete f;
  return mod;
}

void ConfigFile::AddDependency(File *includedby, File *include, ConfigFile *config)
{
  unsigned int i;

  if(!config || config == this)
    return;

  for(i = 0; i < DependsOn.size(); ++i)
    if(DependsOn[i] == config)
    {
      DependsOn[i].AddLink(includedby, include);
      return;
    }

  DependsOn.push_back(CompilerDependency(config));
  DependsOn.back().AddLink(includedby, include);
}

int ConfigFile::Compile(Directory *root)
{
  unsigned int i;
  int depends_built;
  CVector<CString> depend_failed;

  if(!root)
    return 0;

  VALIDATE_POINTER(root)

  CheckDirRec(ObjDir);

  if(Comp)
  {
    VALIDATE_POINTER(Comp)

    #if CMM_ENABLED | SAFEALLOC_ENABLED
      {for(unsigned int i = 0; i < DependsOn.size(); ++i)
        VALIDATE_POINTER(DependsOn[i].Config)}
    #endif

    if(IsLibrary)
      Comp->OutDir = LibDir;
    else
      Comp->OutDir = BinDir;

    CheckDirRec(Comp->OutDir);

    depends_built = 0;

    Comp->BuildIncludeDirs(root);
    Comp->AddAdditonalIncs(AdditonalIncludeDirs);
    Comp->Compile(ObjDir, root, this);

    if(Comp->Errors.size())
      return 0;

    for(i = 0; i < DependsOn.size(); ++i)
      if(DependsOn[i].Config->IsLibrary)
      {
        if(!DependsOn[i].Config->Comp || DependsOn[i].Config->Comp->Errors.size())
        {
          if(DependsOn[i].Config->Comp)
            depend_failed.push_back(DependsOn[i].Config->F);
          else
          {
            CString err;
            err.sprintf("Missing Compiler %s: %s\n", Comp->Name.c_str(), DependsOn[i].Config->F.c_str());
            depend_failed.push_back(err);
          }
          continue;
        }

        if(DependsOn[i].Config->Comp->Built)
          depends_built = 1;
      }

    if(IsLibrary)
      Comp->Link(0);
    else
    {
      if(depend_failed.size())
      {
        printf("\nCannot link. Failed to compile the following dependencies:\n");
        for(i = 0; i < depend_failed.size(); ++i)
          printf("  %s\n", depend_failed[i].c_str());
        return 0;
      }

      for(i = 0; i < AdditionalLibs.size(); ++i)
        Comp->LibsToLink += "\"" + AdditionalLibs[i] + "\" ";

      Comp->Link(depends_built);
    }
    return 1;
  }

  return 0;
}

void ConfigFile::BuildDirectLibList(void)
{
  unsigned int i;

  AdditionalLibs.clear();

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < DependsOn.size(); ++i)
      VALIDATE_POINTER(DependsOn[i].Config)}
  #endif

  for(i = 0; i < DependsOn.size(); ++i)
    if(DependsOn[i].Config->IsLibrary && DependsOn[i].Config->Comp)
      AddAdditionalLib(DependsOn[i].Config->LibDir + DependsOn[i].Config->Comp->FileTitle.c_str());
}

int ConfigFile::BuildIndirectLibList(void)
{
  return BuildIndirectLibList(DependsOn);
}

int ConfigFile::BuildIndirectLibList(const CVector<CompilerDependency> &list)
{
  unsigned int changed = 0;
  unsigned int i, j;


  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < list.size(); ++i)
      VALIDATE_POINTER(list[i].Config)}
  #endif

  for(i = 0; i < list.size(); ++i)
    if(list[i].Config->IsLibrary && list[i].Config->Comp)
      for(j = 0; j < list[i].Config->AdditionalLibs.size(); ++j)
        if(AddAdditionalLib(list[i].Config->AdditionalLibs[j]))
          ++changed;

  return changed;
}

int ConfigFile::AddAdditionalLib(const CString &lib)
{
  unsigned int i;

  if(IsLibrary && Comp && LibDir + Comp->FileTitle == lib)
    return 0;

  VALIDATE_POINTER(Comp)

  for(i = 0; i < AdditionalLibs.size(); ++i)
    if(AdditionalLibs[i] == lib)
      return 0;

  AdditionalLibs.push_back(lib);
  return 1;
}

void ConfigFile::Write(const char *file)
{
  unsigned int i;

  if(Saved)
    return;

  Saved = 1;

  FILE *fp = fopen(file, "wt");

  if(!fp)
  {
    printf("Error: Cannot update %s\n", file);
    return;
  }

  for(i = 0; i < IgnoreDirs.size(); ++i)
    if(IgnoreDirs[i] != ObjDir)
      fprintf(fp, "           SkipDir \"%s\"\n", IgnoreDirs[i].c_str());
  for(i = 0; i < IgnoreFiles.size(); ++i)
    fprintf(fp, "          SkipFile \"%s\"\n", IgnoreFiles[i].c_str());
  for(i = 0; i < AdditonalIncludeDirs.size(); ++i)
    fprintf(fp, "        IncludeDir \"%s\"\n", AdditonalIncludeDirs[i].c_str());
  fprintf(fp, "\n");

  for(i = 0; i < HeaderTypes.size(); ++i)
    fprintf(fp, "     HeaderFileExt \"%s\"\n", HeaderTypes[i].c_str());
  fprintf(fp, "\n");
  for(i = 0; i < SourceTypes.size(); ++i)
    fprintf(fp, "     SourceFileExt \"%s\"\n", SourceTypes[i].c_str());
  fprintf(fp, "\n");

  if(BinDir.size())
    fprintf(fp, "         BinaryDir \"%s\"\n", BinDir.c_str());
  else
    fprintf(fp, "         BinaryDir \"\"\n");

  if(ObjDir.size())
    fprintf(fp, "         ObjectDir \"%s\"\n", ObjDir.c_str());
  else
    fprintf(fp, "         ObjectDir \"\"\n");

  if(LibDir.size())
    fprintf(fp, "        LibraryDir \"%s\"\n", LibDir.c_str());
  else
    fprintf(fp, "        LibraryDir \"\"\n");

  fprintf(fp, "    MinimumRebuild %d\n", MinRebuild != 0);
  fprintf(fp, "           Library %d\n", IsLibrary  != 0);
  fprintf(fp, "\n");

  for(i = 0; i < Comps.size(); ++i)
  {
    fprintf(fp, "          Compiler \"%s\"\n", Comps[i].Name.c_str());
    fprintf(fp, "           Default %d\n",     Comps[i].Default != 0);
    fprintf(fp, "         FileTitle \"%s\"\n", Comps[i].FileTitle.c_str());
    fprintf(fp, "     ShowCompFiles %d\n",     Comps[i].ShowCompFiles != 0);
    fprintf(fp, "AutoIncludeDirTree %d\n",     Comps[i].AutoDirTree != 0);
    fprintf(fp, "  IncludeDirSwitch \"%s\"\n", Comps[i].IncludeDirSwitch.c_str());
    fprintf(fp, "            ObjExt \"%s\"\n", Comps[i].ObjExt.c_str());
    fprintf(fp, "           Compile \"%s\"\n", Comps[i].CompileStr.c_str());
    fprintf(fp, "              Link \"%s\"\n", Comps[i].LinkStr.c_str());
    fprintf(fp, "\n");
  }

  DBase.Write(fp);
  fclose(fp);
}

void ConfigFile::ValidateDirs(void)
{
  ObjDir = DirString(ObjDir).Dir;
  LibDir = DirString(LibDir).Dir;
  BinDir = DirString(BinDir).Dir;
}

void ConfigFile::Read(const char *file)
{
  char *buffer;
  Parser p(file);

  p.Enable(PARSER_HASH            | PARSER_CASE_INSENSITIVE |
           PARSER_CONST_FILE_NAME | PARSER_CONST_TOKEN_SEPS);

  p.AddTokenSeparator(" ",  0);
  p.AddTokenSeparator("\n", 0);
  p.AddTokenSeparator("\r", 0);
  p.AddTokenSeparator("\t", 0);
  p.AddTokenSeparator("//", 0, 1, 0, GenericDiscard);
  p.AddTokenSeparator("/*", 0, 2, 0, GenericDiscard);
  p.AddTokenSeparator("\"\"", 1);
  p.AddTokenSeparator("\"", 0, 3);
  p.AddTokenSeparator("SkipDir",            0, -1, 0, (PCBACK)AddSkipDir,        this);
  p.AddTokenSeparator("SkipFile",           0, -1, 0, (PCBACK)AddSkipFile,       this);
  p.AddTokenSeparator("HeaderFileExt",      0, -1, 0, (PCBACK)AddHeaderType,     this);
  p.AddTokenSeparator("SourceFileExt",      0, -1, 0, (PCBACK)AddSourceType,     this);
  p.AddTokenSeparator("MinimumRebuild",     0, -1, 0, (PCBACK)AddMinimumRebuild, this);
  p.AddTokenSeparator("Library",            0, -1, 0, (PCBACK)AddIsLibrary,      this);
  p.AddTokenSeparator("ObjectDir",          0, -1, 0, (PCBACK)AddObjectDir,      this);
  p.AddTokenSeparator("LibraryDir",         0, -1, 0, (PCBACK)AddLibraryDir,     this);
  p.AddTokenSeparator("BinaryDir",          0, -1, 0, (PCBACK)AddBinaryDir,      this);
  p.AddTokenSeparator("Compiler",           0, -1, 0, (PCBACK)AddCompiler,       this);
  p.AddTokenSeparator("Default",            0, -1, 0, (PCBACK)AddDefault,        this);
  p.AddTokenSeparator("FileTitle",          0, -1, 0, (PCBACK)AddFileTitle,      this);
  p.AddTokenSeparator("ShowCompFiles",      0, -1, 0, (PCBACK)AddShowCompFiles,  this);
  p.AddTokenSeparator("AutoIncludeDirTree", 0, -1, 0, (PCBACK)AutoIncDir,        this);
  p.AddTokenSeparator("IncludeDirSwitch",   0, -1, 0, (PCBACK)IncludeDirSwitch,  this);
  p.AddTokenSeparator("IncludeDir",         0, -1, 0, (PCBACK)AddIncludeDir,     this);
  p.AddTokenSeparator("ObjExt",             0, -1, 0, (PCBACK)AddObjExt,         this);
  p.AddTokenSeparator("Compile",            0, -1, 0, (PCBACK)AddCompileString,  this);
  p.AddTokenSeparator("Link",               0, -1, 0, (PCBACK)AddLinkString,     this);
  p.AddTokenSeparator("[DATABASE]",         0, -1, 0, (PCBACK)LoadDBase,        &DBase);

    // TSet 1 - Handle C++ Comments
  p.AddTokenSet();
  p.AddTokenSeparator("\\\\",   0, -1, 1);
  p.AddTokenSeparator("\\\n\r", 0, -1, 1);
  p.AddTokenSeparator("\\\r\n", 0, -1, 1);
  p.AddTokenSeparator("\\\n",   0, -1, 1);
  p.AddTokenSeparator("\\\r",   0, -1, 1);
  p.AddTokenSeparator("\n",     0,  0);  // Terminate on non escaped new line
  p.AddTokenSeparator("\r",     0,  0);

    // TSet 2 - Handle C Comments
  p.AddTokenSet();
  p.AddTokenSeparator("*/", 0, 0);

    // TSet 3 - Handle Strings
  p.AddTokenSet();
  p.AddTokenSeparator("\\\\", 0, -1, 1); // Ignore \\ .
  p.AddTokenSeparator("\\\"", 0, -1, 1); // Ignore \" within strings
  p.AddTokenSeparator("\"",   0, 0);       // Terminate on a non escaped "

  p.SetTokenSet(0);

  for(buffer = p.GrabToken(); buffer; buffer = p.GrabToken())
    free(buffer);

  for(unsigned int i = 0; i < Comps.size(); ++i)
  {
    Comps[i].Expand();
    if(Comps[i].Default)
      Comp = &Comps[i];
  }

  if(ObjDir.length())
    IgnoreDirs.push_back(ObjDir);
}

char *AddSkipDir(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  config->IgnoreDirs.push_back(buffer);
  free(buffer);
  return str;
}

char *AddSkipFile(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  config->IgnoreFiles.push_back(buffer);
  free(buffer);
  return str;
}

char *AddIncludeDir(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  if(buffer && buffer[0] != '\"')
    config->AdditonalIncludeDirs.push_back(CString(buffer));
  free(buffer);
  return str;
}

char *AddHeaderType(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  config->HeaderTypes.push_back(buffer);
  free(buffer);
  return str;
}

char *AddSourceType(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  config->SourceTypes.push_back(buffer);
  free(buffer);
  return str;
}

char *AddMinimumRebuild(Parser &p, char *str, ConfigFile *config)
{
  VALIDATE_POINTER(config)

  config->MinRebuild = p.GrabInt();
  return str;
}

char *AddIsLibrary(Parser &p, char *str, ConfigFile *config)
{
  VALIDATE_POINTER(config)

  config->IsLibrary = p.GrabInt();
  return str;
}

char *AddObjectDir(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  if(!buffer || buffer[0] == '\"')
    config->ObjDir = "";
  else
    config->ObjDir = buffer;
  free(buffer);
  return str;
}

char *AddLibraryDir(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  if(!buffer || buffer[0] == '\"')
    config->LibDir = "";
  else
    config->LibDir = buffer;
  free(buffer);
  return str;
}

char *AddBinaryDir(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  if(!buffer || buffer[0] == '\"')
    config->BinDir = "";
  else
    config->BinDir = buffer;
  free(buffer);
  return str;
}

char *AddCompiler(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  if(!buffer || buffer[0] == '\"')
    config->Comps.push_back(Compiler("Invalid"));
  else
    config->Comps.push_back(Compiler(buffer));
  free(buffer);
  return str;
}

char *AddDefault(Parser &p, char *str, ConfigFile *config)
{
  int i = p.GrabInt();

  VALIDATE_POINTER(config)

  if(config->Comps.size())
    config->Comps.back().Default = i;
  return str;
}

char *AddFileTitle(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  if(config->Comps.size())
    config->Comps.back().FileTitle = buffer;
  free(buffer);
  return str;
}

char *AutoIncDir(Parser &p, char *str, ConfigFile *config)
{
  int i = p.GrabInt();

  VALIDATE_POINTER(config)

  if(config->Comps.size())
    config->Comps.back().AutoDirTree = i;
  return str;
}

char *IncludeDirSwitch(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  if(buffer && buffer[0] != '\"')
    if(config->Comps.size())
      config->Comps.back().IncludeDirSwitch = buffer;
  free(buffer);
  return str;
}

char *AddShowCompFiles(Parser &p, char *str, ConfigFile *config)
{
  int i = p.GrabInt();

  VALIDATE_POINTER(config)

  if(config->Comps.size())
    config->Comps.back().ShowCompFiles = i;
  return str;
}

char *AddObjExt(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  if(!buffer || buffer[0] == '\"')
    config->Comps.back().ObjExt = ".o";
  else
    config->Comps.back().ObjExt = buffer;
  free(buffer);
  return str;
}

char *AddCompileString(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  if(buffer && buffer[0] != '\"' && config->Comps.size())
    config->Comps.back().CompileStr = buffer;
  free(buffer);
  return str;
}

char *AddLinkString(Parser &p, char *str, ConfigFile *config)
{
  char *buffer = p.GrabToken();

  VALIDATE_POINTER(config)

  if(buffer && buffer[0] != '\"' && config->Comps.size())
    config->Comps.back().LinkStr = buffer;
  free(buffer);
  return str;
}

char *LoadDBase(Parser &p, char *str, Database *dbase)
{
  VALIDATE_POINTER(dbase)

  dbase->Read(p);
  return str;
}
