/* General Structure                                                          *
 *   CMM Page: Each section is the size of a OS Page(PageBlockSize)           *
 *     {Headers....} {Padding} {User Block}{User Block}...                    *
 *                                                                            *
 *   Headers:                                                                 *
 *     For every block, there exists a Header_str. There are 2 special        *
 *       headers in the header table, the Page Pointer and the End of Page.   *
 *       The Page pointer accounts for the allocation of the intial header    *
 *       table, and the End Of Page Accounts for the padding between the      *
 *       header table and the user blocks. In other words, if                 *
 *         PageSize      = 0.5 MB                                             *
 *         PageBlockSize = 4 KB                                               *
 *         NHeadsPerPage = 128                                                *
 *       There are 126 user accessable blocks for a CMM Page. This corresponds*
 *         to header[1] -> header[125].                                       *
 *       If the user request more then the allocated page size, a CMM page    *
 *         will be expanded to fit the needs of the users demand. So, say we  *
 *         allocate a page of 1 MB. Then, header[1] -> header[253] will be    *
 *         user accessable. IE, the Page pointer and the End of Page are a one*
 *         time cost per CMM page.                                            *
 *                                                                            *
 *   Headers Order                                                            *
 *     Page Pointer                                                           *
 *       Len   - Total memory allocated for this page                         *
 *       Start - Start of the user blocks                                     *
 *       Mem   - End of the user blocks                                       *
 *       NHead - Points to the next page                                      *
 *       PHead - Points to the previous page                                  *
 *     User Block Pointer                                                     *
 *     User Block Pointer                                                     *
 *     ...                                                                    *
 *     End Of Page                                                            *
 *       Flags |= END_OF_PAGE                                                 */

#define ERROR_OUTPUT_OVERLOAD_PRINTF 1
#include "MemoryDebugger.h"

#if CMM_ENABLED
#include <windows.h>         /* GetSystemInfo(), VirtualAlloc(), VirtualFree()*/
#include <stdio.h>           /* printf() - Debug only, IE temporary           */
#include <string.h>          /* memset()                                      */

  /* a is in bytes                                                            *
   *   N_BLOCKS_NEEDED(a) - Returns how many user blocks will be used by a    *
   *   N_PAGES_NEEDED(a)  - Returns how many pages are required by a          */
#define N_BLOCKS_NEEDED(a) (((a) + PageBlockSize - 1) / PageBlockSize)
#define N_PAGES_NEEDED(a)  ((N_BLOCKS_NEEDED(a) + NHeadsPerPage - 3) / \
                            (NHeadsPerPage - 2))

  /* Header_str Flags defines                                                 */
#define FREE        (1 << 0) /* Memory is currently free                      */
#define HAS_PADDING (1 << 1) /* Allocated mem has a page of padding following */
#define END_OF_PAGE (1 << 7) /* Header denotes the end of the page            */

struct Header_str
{
  unsigned int       Flags;  /* See defines above                             */
  unsigned int       Len;    /* Number of bytes in the block                  */
  unsigned char     *Start;  /* Pointer to the start of the block             */
  unsigned char     *Mem;    /* Pointer to the start of the user mem          */
  struct Header_str *Prev;   /* Free List Previous Pointer                    */
  struct Header_str *Next;   /* Free List Next Pointer                        */
  struct Header_str *NHead;  /* Next Header                                   */
  struct Header_str *PHead;  /* Previous Header                               */
};

typedef struct Header_str Header_str;

static int          GetPageSize     (void);
static int          EnableReadWrite (Header_str  *block);
static int          DisableReadWrite(Header_str  *block);
static Header_str  *FindHeader      (void        *mem);
static void         FragmentBlock   (unsigned int bytes, Header_str *head);
static int          AllocPage       (unsigned int bytes);
static void         RemoveBlock     (Header_str  *block);
static void         PushBlock       (Header_str  *block);
static void         PrintSize       (unsigned int size);

static unsigned int PageSize      = 0; /* Full size of a CMM page             */
static unsigned int PageBlockSize = 0; /* Size of each user block/OS Page size*/
static unsigned int NHeadsPerPage = 0; /* Number of headers per page          */
static Header_str  *FreeList      = 0; /* Free memory blocks                  */
static Header_str  *Page          = 0; /* Pointer to the top of page list     */
static unsigned int AddPadding    = 1; /* Add padding to allocated mem?       */
static unsigned int MinPageSize   = 10 * 1024 * 1024; /* Reserve 10 MB per pag*/

void *CMMmalloc(unsigned int bytes)
{
  Header_str *block = FreeList;

  if(!bytes)
    return 0;

  if(PageSize == 0)
    GetPageSize();

  if(AddPadding)
    bytes += PageBlockSize;

  for(; block; block = block->Next)
    if(block->Len >= bytes)
      break;

  if(!block)
  {
    if(!AllocPage(bytes))
      return 0;
    else
    {
      if(AddPadding)
        bytes -= PageBlockSize;
      return CMMmalloc(bytes);
    }
  }

  RemoveBlock(block);

  if((block->Len - bytes) > PageBlockSize)
    FragmentBlock(bytes, block);

  block->Mem = block->Start + (block->Len - bytes);

  if(!EnableReadWrite(block))
    return 0;

  return block->Mem;
}

void CMMfree(void *mem)
{
  Header_str *head;
  if(!mem)
    return;

  head = FindHeader(mem);
  if(!head)
    return;

  if(!DisableReadWrite(head))
    return;

    /* Merge Left */
  if(head->PHead->Flags & FREE)
  {
    head->PHead->Len += head->Len;

    head->PHead->NHead = head->NHead;
    head->NHead->PHead = head->PHead;

    head = head->PHead;
  }
  else
    PushBlock(head);

    /* Merge Right */
  if(head->NHead->Flags & FREE)
  {
    RemoveBlock(head->NHead);
    head->Len += head->NHead->Len;

    head->NHead->NHead->PHead = head;
    head->NHead = head->NHead->NHead;
  }
}

void *CMMrealloc(void *mem, unsigned int bytes)
{
  unsigned int blocksize;
  void *new_mem;
  Header_str *head;

  if(!mem)
    return CMMmalloc(bytes);

  if(!bytes)
  {
    CMMfree(mem);
    return 0;
  }

  head = FindHeader(mem);

  if(!head)
    return 0;

  blocksize = (unsigned int)((head->Start + head->Len) - head->Mem);
  if(head->Flags & HAS_PADDING)
    blocksize -= PageBlockSize;

    /* If the memry does not need to be resized, and we don't need to append  *
     *   or remove padding, just return the mem.                              */
  if(blocksize == bytes)
    if(( (head->Flags & HAS_PADDING) &&  AddPadding) ||
       (!(head->Flags & HAS_PADDING) && !AddPadding))
      return mem;

  /* I'm to lazy to optimize this, plus there are a lot of cases. So, just    *
   *   allocate a new block, copy everything, and return.                     */

  if(bytes < blocksize)
    blocksize = bytes;
  new_mem = CMMmalloc(bytes);
  memcpy(new_mem, mem, blocksize);
  CMMfree(mem);

  return new_mem;
}

void *CMMcalloc(unsigned int elements, unsigned int elementsize)
{
  void *mem = CMMmalloc(elements * elementsize);

  if(mem)
    memset(mem, 0, elements * elementsize);

  return mem;
}

void CMMEnablePadding(void)
{
  AddPadding = 1;
}

void CMMDisablePadding(void)
{
  AddPadding = 0;
}

void CMMMinPageSize(unsigned int bytes)
{
  MinPageSize = bytes;
}

unsigned int CMMReserve(unsigned int bytes)
{
  return AllocPage(bytes);
}

void CMMDumpReport(void)
{
  printf("Current Allocation Status\n<--------------->\n");
  printf("Active Pointers: %d\n" ,CMMGetNPointers());
  printf("Active Bytes   : "); PrintSize(CMMGetUserSize()); printf("\n");
  printf("----\n");
  printf("Padding Size   : "); PrintSize(CMMGetPaddingSize()); printf("\n");
  printf("Total Alloc    : "); PrintSize(CMMGetPaddingSize() + CMMGetUserSize()); printf("\n");
  printf("<--------------->\n\n");
}

void CMMDumpBlocks(char *spaces)
{
  Header_str *page = Page;
  Header_str *head;
  int i = 0;

  for(; page; page = page->NHead, ++i)
  {
    printf("%sPage: %2d (%p)\n", spaces, i, (void *)page);
    printf("%s  Page Header: %p\n", spaces, (void *)page);
    printf("%s          Len: ", spaces); PrintSize((int)page->Len);printf("\n");
    printf("%s   User Range: 0x%p -> 0x%p\n", spaces, page->Start, page->Mem);
    printf("%s   Max Blocks: %u\n", spaces,
                       (unsigned int)(page->Mem - page->Start) / PageBlockSize);
    printf("\n");

    head = page + 1;
    for(; !(head->Flags & END_OF_PAGE); head = head->NHead)
    {
      printf("%s  Block: %p (Mem: %p)",spaces,(void *)head,(void *)head->Start);
      if(head == FreeList)
        printf("  <-- FreeList List");
      printf("\n");
      printf("%s       Len: ", spaces); PrintSize((int)head->Len); printf("\n");
      printf("%s     Flags: ", spaces);

      if(head->Flags & FREE)
        printf("|Free Block|\n");
      else
      {
        printf("|User Data | (");
        if(head->Flags & HAS_PADDING)
        {
          PrintSize((int)((head->Start + head->Len) - head->Mem)-PageBlockSize);
          printf(" -- ");
          PrintSize(PageBlockSize);
          printf(" of padding");
        }
        else
          PrintSize((int)((head->Start + head->Len) - head->Mem));
        printf(" -- %p)\n", head->Mem);
      }

      if(head->Flags & FREE)
      {
        printf("%s      Next: %p\n", spaces, (void *)head->Next);
        printf("%s      Prev: %p\n", spaces, (void *)head->Prev);
      }

      printf("\n");
    }
  }
}

void CMMDumpMemoryMap(void *mem)
{
  Header_str *page = Page;
  Header_str *head;
  int i = 0;
  unsigned int j;
  unsigned int len;

  for(; page; page = page->NHead, ++i)
  {
    printf("Page: %2d (%p)\n", i, (void *)page);
    head = page + 1;
    for(; !(head->Flags & END_OF_PAGE); head = head->NHead)
    {
      if(mem && mem != head->Mem)
        continue;

      if(head->Flags & FREE)
      {
        printf("%p: ", head->Start);
        PrintSize(head->Len);
        printf(": FreeMemory\n\n");
      }
      else
      {
        len = CMMPointerSize(head->Mem);
        printf("%p: ", head->Mem);
        PrintSize(len);
        printf("\n");
        for(j = 0; j < len; ++j)
        {
          if(j && !(j % 20))
            printf("\n");
          if(!(j % 20))
            printf("%p", head->Mem + j);
          if(!(j % 4))
            printf(": ");
          printf("%02X ", head->Mem[j]);
        }
        printf("\n\n");
      }
    }
  }
}

unsigned int CMMIsAValidPointer(void *_mem)
{
  unsigned int i;
  unsigned int pad;
  Header_str *page = Page;
  unsigned char *mem = (unsigned char *)_mem;

  for(; page; page = page->NHead)
    if(page->Start <= mem && mem < page->Mem)
    {
      i = (unsigned int)(mem - page->Start) / PageBlockSize + 1; /* +1 for Page Header */

      if(page[i].Flags & FREE)
        return 0;

      if(page[i].Flags & HAS_PADDING)
        pad = PageBlockSize;
      else
        pad = 0;

      if(page[i].Mem <= mem && mem < page[i].Start + page[i].Len - pad)
        return 1;

      return 0;
    }

  return 0;
}

unsigned int CMMIsABasePointer(void *mem)
{
  Header_str *head = FindHeader(mem);
  if(head && !(head->Flags & FREE))
    return 1;
  return 0;
}

unsigned int CMMPointerSize(void *mem)
{
  Header_str *head = FindHeader(mem);
  if(!head || head->Flags & FREE)
    return 0;

  if(head->Flags & HAS_PADDING)
    return (unsigned int)(head->Start + head->Len - head->Mem - PageBlockSize);
  return (unsigned int)(head->Start + head->Len - head->Mem);
}

unsigned int CMMGetUserSize(void)
{
  unsigned int bytes = 0;
  Header_str *page = Page;
  Header_str *head;

  for(; page; page = page->NHead)
    for(head = page + 1; !(head->Flags & END_OF_PAGE); head = head->NHead)
      if(!(head->Flags & FREE))
      {
        bytes += (unsigned int)((head->Start + head->Len) - head->Mem);
        if(head->Flags & HAS_PADDING)
          bytes -= PageBlockSize;
      }

  return bytes;
}

unsigned int CMMGetPaddingSize(void)
{
  unsigned int bytes = 0;
  Header_str *page = Page;
  Header_str *head;

  for(; page; page = page->NHead)
  {
    bytes += PageBlockSize;
    for(head = page + 1; !(head->Flags & END_OF_PAGE); head = head->NHead)
      if(!(head->Flags & FREE))
      {
        bytes += (unsigned int)(head->Mem - head->Start);
        if(head->Flags & HAS_PADDING)
          bytes += PageBlockSize;
      }
  }

  return bytes;
}

unsigned int CMMGetFreeSize(void)
{
  unsigned int bytes = 0;
  Header_str *page = Page;
  Header_str *head;

  for(; page; page = page->NHead)
    for(head = page + 1; !(head->Flags & END_OF_PAGE); head = head->NHead)
      if(head->Flags & FREE)
        bytes += head->Len;

  return bytes;
}

unsigned int CMMGetOverhead(void)
{
  unsigned int bytes = 0;
  Header_str *page = Page;

  for(; page; page = page->NHead)
    bytes += (unsigned int)((unsigned char *)page->Start - (unsigned char *)page-PageBlockSize);

  bytes += CMMGetPaddingSize();

  return bytes;
}

unsigned int CMMGetReserved(void)
{
  unsigned int bytes = 0;
  Header_str *page = Page;

  for(; page; page = page->NHead)
    bytes += page->Len;

  return bytes;
}

float CMMGetUsagePercent(void)
{
  return (float)(CMMGetUserSize() + CMMGetOverhead()) / CMMGetReserved();
}

unsigned int CMMGetNPointers(void)
{
  unsigned int ptrs = 0;
  Header_str *page = Page;
  Header_str *head;

  for(; page; page = page->NHead)
    for(head = page + 1; !(head->Flags & END_OF_PAGE); head = head->NHead)
      if(!(head->Flags & FREE))
        ++ptrs;

  return ptrs;
}

int GetPageSize(void)
{
  SYSTEM_INFO info;
  GetSystemInfo(&info);
  PageBlockSize = (unsigned int)info.dwPageSize;

  NHeadsPerPage = PageBlockSize / sizeof(Header_str);

  PageSize = NHeadsPerPage * PageBlockSize;

  return (int)PageSize;
}

int EnableReadWrite(Header_str *block)
{
  unsigned int len = block->Len;

  if(AddPadding)
  {
    len -= PageBlockSize;
    block->Flags |= HAS_PADDING;
  }

  if(VirtualAlloc(block->Start, len, MEM_COMMIT, PAGE_READWRITE))
    return 1;
  return 0;
}

int DisableReadWrite(Header_str *block)
{
  if(VirtualFree(block->Start, block->Len, MEM_DECOMMIT))
    return 1;
  return 0;
}

Header_str *FindHeader(void *_mem)
{
  unsigned int i;
  Header_str *page = Page;
  unsigned char *mem = (unsigned char *)_mem;

  for(; page; page = page->NHead)
    if(page->Start <= mem && mem < page->Mem)
    {
      i = (unsigned int)(mem - page->Start) / PageBlockSize + 1; /* +1 for Page Header */
      if(page[i].Mem == mem)
        return page + i;
      return 0;
    }

  return 0;
}

void FragmentBlock(unsigned int bytes, Header_str *head)
{
  Header_str *nblock;
  int nblocksneeded = N_BLOCKS_NEEDED(bytes);
  int nblockshave   = N_BLOCKS_NEEDED(head->Len);

  if(nblocksneeded == nblockshave)
    return;

  nblock = head + nblocksneeded;
  head->Len = nblocksneeded * PageBlockSize;
  head->NHead->PHead = nblock;

  nblock->NHead = head->NHead;
  nblock->PHead = head;

  head->NHead   = nblock;

  nblock->Len   = (nblockshave - nblocksneeded) * PageBlockSize;

  PushBlock(nblock);
}

int AllocPage(unsigned int bytes)
{
  int i;
  int npages;
  int nuserblocks;
  Header_str *head;

  if(!NHeadsPerPage)
    if(!GetPageSize())
      return 0;

  if(bytes < MinPageSize)
    bytes = MinPageSize;

  npages = N_PAGES_NEEDED(bytes);
  nuserblocks = NHeadsPerPage * npages - npages - 1;

  head =(Header_str *)VirtualAlloc(0,PageSize*npages,MEM_RESERVE,PAGE_NOACCESS);
  if(!head)
    return 0;

    /* VirutalAlloc succeeded, however, we cannot access any of the memory.   */
    /* Commit the required number of pages for the header table.              */
  head = (Header_str *)VirtualAlloc(head, PageBlockSize * npages, MEM_COMMIT,
                                                                PAGE_READWRITE);
  if(!head)
    return 0;

    /* Set up the Page Pointer */
  head[0].Flags = 0;
  head[0].Len   = PageSize * npages;
  head[0].Start = (unsigned char *)head + PageBlockSize * (npages + 1);
  head[0].Mem   = (unsigned char *)head + PageSize * npages;
  head[0].Next  = 0;
  head[0].Prev  = 0;
  head[0].NHead = Page;
  head[0].PHead = 0;
  if(Page)
    Page->PHead = head;
  Page = head;

  for(i = 0; i < nuserblocks; ++i)
    head[i + 1].Start = head[0].Start + PageBlockSize * i;

    /* Set up the first user block header */
  memset(head + 1, 0, sizeof(Header_str));
  head[1].Len    = (unsigned int)(head[0].Mem - head[0].Start);
  head[1].Start  = head[0].Start;
  head[1].PHead  = head;
  head[1].NHead  = head + npages + nuserblocks;

  PushBlock(head + 1);

    /* set up the End Of Page header */
  head[1].NHead->PHead = head + 1;
  head = head[1].NHead;
  head->Flags = END_OF_PAGE;

  return 1;
}

void RemoveBlock(Header_str *block)
{
  block->Flags &= ~FREE;
  if(block == FreeList)
    FreeList = block->Next;

  if(block->Next)
    block->Next->Prev = block->Prev;

  if(block->Prev)
    block->Prev->Next = block->Next;
}

void PushBlock(Header_str *block)
{
  block->Flags &= ~HAS_PADDING;
  block->Flags |= FREE;

  if(FreeList)
    FreeList->Prev = block;

  block->Next = FreeList;
  block->Prev = 0;

  FreeList = block;
}

void PrintSize(unsigned int size)
{
  unsigned int gb;
  unsigned int mb;
  unsigned int kb;
  unsigned int  b;

  gb =  size / 1073741824;
  mb = (size % 1073741824) / 1048576;
  kb = (size % 1048576) / 1024;
  b  =  size % 1024;

       if(gb) printf("%4u.%02u GB", gb, (int)(mb /1024.0 * 100));
  else if(mb) printf("%4u.%02u MB", mb, (int)(kb /1024.0 * 100));
  else if(kb) printf("%4u.%02u KB", kb, (int)( b /1024.0 * 100));
  else        printf("%4u Bytes", b);
}

#ifdef __cplusplus
#if CMM_OVERLOAD_CPP_ALLOC
void *operator new(unsigned int size) throw(std::bad_alloc)
{
  if(!size)  size = sizeof(int);

  void *ret = CMMmalloc(size);
  
  if(!ret)
    throw(std::bad_alloc());
    
  return ret;
}

void *operator new(unsigned int size, const std::nothrow_t&) throw()
{
  if(!size) size = sizeof(int);
  return CMMmalloc(size);
}

void *operator new[](unsigned int size) throw(std::bad_alloc)
{
  if(!size) size = sizeof(int);

  void *ret = CMMmalloc(size);
  
  if(!ret)
    throw(std::bad_alloc());
    
  return ret;
}

void *operator new[](unsigned int size, const std::nothrow_t&) throw()
{
  if(!size) size = sizeof(int);
  
  return CMMmalloc(size);
}

void  operator delete(void * ptr) throw()
{
  CMMfree(ptr);
}

void  operator delete(void * ptr, const std::nothrow_t&) throw()
{
  CMMfree(ptr);
}

void  operator delete[](void * ptr) throw()
{
  CMMfree(ptr);
}

void  operator delete[](void * ptr, const std::nothrow_t&) throw()
{
  CMMfree(ptr);
}

#endif /* CMM_OVERLOAD_CPP_ALLOC */
#endif /* __cplusplus */
#endif /* CMM_ENABLED */
