/*******************************************************************************
CMMmalloc         - CMM version of malloc()
CMMrealloc        - CMM version of realloc()
CMMcalloc         - CMM version of calloc()
CMMfree           - CMM version of free()

  Configuration Settings
CMMEnablePadding  - Enables padding support. This will cause the CMM to append
                      a page of no access padding to the end of each pointer.
                      This allows for overflow detection. By default, this is
                      enabled.
CMMDisablePadding - Disables the above. No padding pages will be appended to
                      the end of a allocation.
CMMMinPageSize    - Sets the minimum CMM page size. This defaults to 10 MB.
                      Decreasing this value will increase the number of pages
                      that must be allocated, and increase the amount of time
                      it takes to free()/delete a pointer. A high value (~100MB)
                      is recommended if performance matters.
CMMReserve        - Allocates a new CMM page with the amount of memory
                      specified. This number will be rounded up to the next
                      CMM page size. A 1 will be returned if this functions
                      succeeded. Otherwise, a 0 will be returned.

  Extra Debug Related Functions
CMMDumpReport     - Prints out a basic allocation report
CMMDumpBlocks     - Prints out all pages, and all page headers
CMMDumpMemoryMap  - Prints out all allocated memory. If a valid pointer is
                      passed to the function, only that pointer's memory
                      will be dumped.
CMMIsABasePointer - Returns 1 if the pointer points to the start of a user
                      allocated memory block.
CMMIsAValidPointer- Returns 1 if the pointer is within a user allocated memory
                      block.
CMMPointerSize    - Returns the size of the block of memory pointed. If the
                      pointer is not a base pointer, 0 will be returned.
CMMGetUserSize    - Returns the amount of memory the user has requested.
CMMGetPaddingSize - Returns the amount of memory used as padding
CMMGetFreeSize    - Returns the amount of memory available, based on the current
                      reserves.
CMMGetOverhead    - Returns how much memory is being used to manage all
                      pointers, including the padding bytes.
CMMGetReserved    - Returns how much memory is currently reserved.
CMMGetUsagePercent- (GetUserSize() + GetOverhead()) / GetReserved()
CMMGetNPointers   - Returns the total number of active user pointers
*******************************************************************************/

#ifndef CMM_H
#define CMM_H

  /* Set to the default defines, if they were not defined externally */
#ifndef CMM_ENABLED
  #define CMM_ENABLED            1
#endif
#ifndef CMM_OVERLOAD_C_ALLOC
  #define CMM_OVERLOAD_C_ALLOC   1
#endif
#ifndef CMM_OVERLOAD_CPP_ALLOC
  #define CMM_OVERLOAD_CPP_ALLOC 1
#endif

#if CMM_ENABLED

#include <stdlib.h>

#if CMM_OVERLOAD_C_ALLOC
  #define  malloc(size     ) CMMmalloc(size)
  #define  calloc(ele, size) CMMcalloc(ele, size)
  #define realloc(ptr, size) CMMrealloc(ptr, size)
  #define    free(ptr      ) CMMfree(ptr)
#endif

void        *CMMmalloc (unsigned int   bytes                             );
void        *CMMrealloc(         void *mem,      unsigned int bytes      );
void        *CMMcalloc (unsigned int   elements, unsigned int elementsize);
void         CMMfree   (         void *mem                               );

  /* Configuration Settings */
void         CMMEnablePadding  (void);     /* Add padding after each alloc    */
void         CMMDisablePadding (void);     /* Do not add padding after allco  */
void         CMMMinPageSize    (unsigned int bytes); /* Set min page size     */
unsigned int CMMReserve        (unsigned int bytes); /* Reserve some memory   */

  /* Extra Debug Related Functions */
void         CMMDumpReport     (void);     /* Write out a general mem report  */
void         CMMDumpBlocks     (char *spc);/* Prints pages and mem info       */
void         CMMDumpMemoryMap  (void *mem);/* 0 to dump all memory            */
unsigned int CMMIsABasePointer (void *mem);/* 1 if a base memory pointer      */
unsigned int CMMIsAValidPointer(void *mem);/* 1 if base or in valid range     */
unsigned int CMMPointerSize    (void *mem);/* 0 on error, else memory size    */
unsigned int CMMGetUserSize    (void);     /* Total user allocated memory     */
unsigned int CMMGetPaddingSize (void);     /* Total memory used for padding   */
unsigned int CMMGetFreeSize    (void);     /* Free space, based on reserved   */
unsigned int CMMGetOverhead    (void);     /* Padding size + Allocation Tables*/
unsigned int CMMGetReserved    (void);     /* Total memory the manager has    */
float        CMMGetUsagePercent(void);     /* Used memory / Reserved          */
unsigned int CMMGetNPointers   (void);     /* Total number of allocations     */

#ifdef __cplusplus
  #if CMM_OVERLOAD_CPP_ALLOC
    #include <new>
      /* Disable the warning that the compiler ignores the throw type         */
    #pragma warning(disable : 4290)

    void *operator new     (unsigned int size) throw(std::bad_alloc);
    void *operator new     (unsigned int size, const std::nothrow_t&) throw();
    void *operator new[]   (unsigned int size) throw(std::bad_alloc);
    void *operator new[]   (unsigned int size, const std::nothrow_t&) throw();
    void  operator delete  (void * ptr)                               throw();
    void  operator delete  (void * ptr, const std::nothrow_t&)        throw();
    void  operator delete[](void * ptr)                               throw();
    void  operator delete[](void * ptr, const std::nothrow_t&)        throw();
  #endif /* CMM_OVERLOAD_CPP_ALLOC */
#endif   /* __cplusplus            */
#endif   /* CMM_ENABLED            */
#endif   /* CMM_H                  */
