/*******************************************************************************
filename      SafeAlloc.cpp
author        Craig Williams
e-mail        cwilliam@digipen.edu

Brief Description:
  Notice: SafeAlloc.h must be included in every .c/cpp file. If you do not,
            and the memory debugger is enabled, the heap may become corrupted
            since calls to *alloc() and free() will not be intercepted in the
            files that do not include SafeAlloc.h.

  Provides various memory debugging functions. Including this header will
    overwrite *alloc(), free(), new and delete. When you allocate some memory,
    a specified number of padding bytes will be added to each end of the memory.
    These padding blocks will have a signature written to them. If the signature
    changes, you will be notifed.

  Functions:
    s*alloc()  - Overloaded version of the specified alloc()
    sn*alloc() - Same as above; however, you can attach a meanful name to the
                   memory.
    sfree()    - Overloaded version of free()

  Memory Profiling Functions:
    sRenameAlloc()   - Allows you to attach a meanful name to a pointer. This
                         name will be printed out with the pointer when ever you
                         call a Dump*().
    ProtectAlloc()   - Marks a pointer as "protected". This prevents
                         GarbageCollect() from freeing the data when called.
                         This is also used to signify that the specfied pointer
                         is suposed to be there.
    UnProtectAlloc() - Removes the "protected" flag.
    ProtectAll()     - Sets all the active poitners as protected.
    GarbageCollect() - Removes all memory that was not marked as protected. This
                         function should not be used.

  Memory Status Functions:
    DumpAllocStatus()     - Prints out how much memory is being used, as well
                              as how many pointers there are. Calling this
                              function at the end of main() or similar will
                              give you a report of the amount of memory was
                              leaked.
    DumpAllocatedMem()    - Lists all the pointers that the memory debugger is
                              currently tracking includeing the alloc location,
                              size, and a symbol name if it has one.
    DumpAllocatedMemMin() - Same as above; however, the output is far more
                              sreamlined. IE, 1-2 lines per pointer.
    DumpAllocatedMemNonProtected() - Same as DumpAllocatedMem(), but it only
                                       dumps non protected pointers.
    CheckAllPadding()     - Checks every pointer for over/under writes. This is
                              done by checking the padding area attached to each
                              pointer.

  Private Functions:
    GetAllocIndex()    - Attempts to locate the specfied pointer in the
                           allocation list. If the pointer is not found, a -1
                           will be returned. Else, the pointers index is
                           returned.
    PrintSize()        - Prints out the specifed number of bytes in easily
                           readable form. IE, 2048 bytes will be printed out as
                           2 KB instead.
    MarkPadding()      - Marks the padding of a block of memory with the
                           specified signature.
    CheckPadding()     - Checks the padding block attached to memory. If the
                           signatures do not match, a error message will be
                           printed to stdout
    CheckOldRecords()  - Scans the pointer list for the specified pointer. This
                           will only check the pointers that have been
                           previously freed.
    PrintPointerData() - Prints out the pointer information in a nice
                           streamlined form.
*******************************************************************************/
#define SAFEALLOC_OVERLOAD_C_ALLOC   0
#define CMM_OVERLOAD_C_ALLOC         1
#define ERROR_OUTPUT_OVERLOAD_PRINTF 1
#include "MemoryDebugger.h"

#if SAFEALLOC_ENABLED

#include <stdio.h>
#include <stdlib.h>

#if DEBUG_WINDOWS
    /* Disable Pointer -> unsigned int. This is only done to generate the hash*
     *   This is safe. Once the hash is generated, the pointers are compared  *
     *   to find the correct slot.                                            */
  #pragma warning(disable : 4311)
#endif

#define UNOCCUPIED   0   /* Memoery Hashing state defines. Should not be      */
#define OCCUPIED     1   /*   modified.                                       */
#define DELETED      2   /*                                                   */

#define FROM_MALLOC  0
#define FROM_FREE    0
#define FROM_NEW     1
#define FROM_DEL     1
#define FROM_NEW_A   2
#define FROM_DEL_A   2
#define FROM_REALLOC 3

static void PrintFromMiss   (int i, char *file, int linenum, int from);
static int  GetAllocIndex   (void *ptr);
static int  CheckOldRecords (void *ptr);
static void PrintSize       (int size);
static void MarkPadding     (void *ptr, int size);
static void CheckPadding    (void *ptr);
static void PrintPointerData(int i);
static int  MHInsert        (unsigned long Key);
static void MHRemove        (unsigned long Key);
static int  MHFind          (unsigned long Key);
static int  MHFindDeleted   (unsigned long Key);

void CheckAllPadding(void);

struct Alloc_str
{
  unsigned int Size;
  void *Ptr;
  char *Name;
  char *File;
  char *FreeFile;
  int   FreeLine;
  int   Line;
  int   Active;
  int   Protected;
  int   Flags;
};

struct MHSlot_str
{
  unsigned long Key;
  struct Alloc_str Data;
  int State;
};

struct MemHash_str
{
  struct MHSlot_str *Table;
  unsigned int  TableSize;
  unsigned int  Elements;
  float     Load;
  float     MaxLoad;
  float     Growth;
};

static struct MemHash_str *MemHash  = 0;
#if DEBUG_WINDOWS
  static struct MemHash_str *AddrHash = 0;
  static int BreakOnError = 1;
#endif
/*******************************************************************************
   Function: smalloc()

Description: Safe Malloc. Allocates the specifed amount of memory (in bytes)

     Inputs: size - How much memory to allocate (in bytes)

    Outputs: Pointer to the new chunk of memory
*******************************************************************************/
void *smalloc(unsigned int size, char *file, int linenum)
{
  return snmalloc(size, 0, file, linenum, FROM_MALLOC);
}

/*******************************************************************************
   Function: snmalloc()

Description: Safe Name Malloc. Allocates the specifed amount of memory
              (in bytes), and then attach a name to it.

     Inputs: size  - How much memory to allocate (in bytes)
             *name - Pointer to a string containing the identification name

    Outputs: Pointer to the new chunk of memory
*******************************************************************************/
void *snmalloc(unsigned int size, char *name, char *file, int linenum, int from)
{
  void *ptr;
  int i;

  if(size == 0)
    return 0;

  ptr = malloc(size + PADDING_BEFORE + PADDING_AFTER);

  if(!ptr)
    return 0;

  i = MHInsert((unsigned long)ptr);

  if(i == -1)
    return 0;

  MemHash->Table[i].Data.Ptr      = ptr;
  MemHash->Table[i].Data.Name     = name;
  MemHash->Table[i].Data.File     = file;
  MemHash->Table[i].Data.Line     = linenum;
  MemHash->Table[i].Data.FreeFile = 0;
  MemHash->Table[i].Data.FreeLine = 0;
  MemHash->Table[i].Data.Size     = size;
  MemHash->Table[i].Data.Active   = 1;
  MemHash->Table[i].Data.Flags    = from;

  MarkPadding(MemHash->Table[i].Data.Ptr, size);

  return (unsigned char *)MemHash->Table[i].Data.Ptr + PADDING_BEFORE;
}

/*******************************************************************************
   Function: scalloc()

Description: Allocates a new chuck of memory, and then initalizes it to 0

     Inputs: esize - Size of each element in the array
             noe   - Number of Elements. How many elements to allocate

    Outputs: Pointer to the new chunk of memory.
*******************************************************************************/
void *scalloc(unsigned int esize, unsigned int noe, char *file, int linenum)
{
  return sncalloc(esize, noe, 0, file, linenum);
}

/*******************************************************************************
   Function: sncalloc()

Description: Safe Name Calloc. Allocates the specifed amount of memory
              (in bytes), and then attach a name to it.

     Inputs: esize - Element size. Size of each element
             noe   - Number of Elements. How many elements to allocate
             *name - Pointer to a string containing the identification name

    Outputs: Pointer to the new chunk of memory
*******************************************************************************/
void *sncalloc(unsigned int esize, unsigned int noe, char *name, char *file, int linenum)
{
  void *ptr;
  int size = esize * noe;
  int i;

  if(size == 0)
    return 0;

  ptr = calloc(1, size + PADDING_BEFORE + PADDING_AFTER);

  if(!ptr)
    return 0;

  i = MHInsert((unsigned long)ptr);

  if(i == -1)
    return 0;

  MemHash->Table[i].Data.Ptr      = ptr;
  MemHash->Table[i].Data.Name     = name;
  MemHash->Table[i].Data.File     = file;
  MemHash->Table[i].Data.Line     = linenum;
  MemHash->Table[i].Data.FreeFile = 0;
  MemHash->Table[i].Data.FreeLine = 0;
  MemHash->Table[i].Data.Size     = size;
  MemHash->Table[i].Data.Active   = 1;
  MemHash->Table[i].Data.Flags    = FROM_MALLOC;

  MarkPadding(MemHash->Table[i].Data.Ptr, size);

  return (unsigned char *)MemHash->Table[i].Data.Ptr + PADDING_BEFORE;
}

/*******************************************************************************
   Function: srealloc()

Description: Reallocates a chunk of memory. If the pointer was not crated by
                this funtion, a new chunk of memory will be allocated,

     Inputs: *ptr - Pointer to the chunk of memory to resize
             size - New size of the chunk of memory

    Outputs: Pointer to the new chunk of memory
*******************************************************************************/
void *srealloc(void *ptr, unsigned int size, char *file, int linenum)
{
  return snrealloc(ptr, size, 0, file, linenum);
}

/*******************************************************************************
   Function: snrealloc()

Description: Safe Name realloc. Reallocates the specifed amount of memory
              (in bytes), and then attach a name to it.

     Inputs: *ptr  - Pointer to the chunk of memory to resize.
             size  - New size of the chunk of memory
             *name - Pointer to a string containing the identification name

    Outputs: Pointer to the new chunk of memory
*******************************************************************************/
void *snrealloc(void *ptr, unsigned int size, char *name, char *file, int linenum)
{
  int i = GetAllocIndex(ptr);

  if(i == -1)
    return snmalloc(size, name, file, linenum, FROM_MALLOC);

  if(size == 0)
  {
    sfree(ptr, file, linenum, FROM_MALLOC);
    return 0;
  }

  CheckPadding(ptr);

  if(MemHash->Table[i].Data.Flags != FROM_MALLOC)
  {
    PrintFromMiss(i, file, linenum, FROM_REALLOC);
    return 0;
  }

  MHRemove((unsigned long)((unsigned char *)ptr - PADDING_BEFORE));
  ptr = realloc(MemHash->Table[i].Data.Ptr, size + PADDING_BEFORE + PADDING_AFTER);

  if(!ptr)
    return 0;

  i = MHInsert((unsigned long)ptr);

  if(i == -1)
    return 0;

  MemHash->Table[i].Data.Ptr      = ptr;
  MemHash->Table[i].Data.Name     = name;
  MemHash->Table[i].Data.File     = file;
  MemHash->Table[i].Data.Line     = linenum;
  MemHash->Table[i].Data.FreeFile = 0;
  MemHash->Table[i].Data.FreeLine = 0;
  MemHash->Table[i].Data.Size     = size;
  MemHash->Table[i].Data.Active   = 1;
  MemHash->Table[i].Data.Flags    = FROM_MALLOC;

  MarkPadding(MemHash->Table[i].Data.Ptr, size);

  return (unsigned char *)MemHash->Table[i].Data.Ptr + PADDING_BEFORE;
}

/*******************************************************************************
   Function: sfree()

Description: Frees a chunk memory previously in use

     Inputs: *ptr - Pointer to the chunk of memory to free

    Outputs: Number of bytes freed
*******************************************************************************/
int sfree(void *ptr, char *file, int linenum, int from)
{
  int i = GetAllocIndex(ptr);

  #if ACTIVECHECK == 1
    CheckAllPadding();
  #endif

  if(i == -1)
  {
    if(file || linenum == -1)
    {
      printf("ERROR: Attempting to free something we didnt allocate or already freed:%p\n", ptr);

      if(file)
      {
        printf("       Generated at:\n");
        printf("         File: %s\n", file);
        printf("         Line: %d\n", linenum);
      }

      i = CheckOldRecords(ptr);

      if(i != -1)
      {
        if(MemHash->Table[i].Data.FreeFile)
        {
          printf("       Originally deleted at:\n");
          printf("         File: %s\n", MemHash->Table[i].Data.FreeFile);
          printf("         Line: %d\n", MemHash->Table[i].Data.FreeLine);
        }
        printf("       Originally Allocated at:\n");
        printf("         ");
        PrintPointerData(i);
      }

      #if FORCEFREE == 1
        printf("       FORCEFREE enabled. calling free(%p);\n", ptr);
        free(ptr);
      #endif

      printf("\n");
    }

    #if DEBUG_WINDOWS
        /* You arrived here if there was a double delete or you freed garbage.*
         *    See the console or output file for more details.                */
      if(BreakOnError) __asm int 3;
    #endif
    return 0;
  }

  if(MemHash->Table[i].Data.Flags != from)
  {
    PrintFromMiss(i, file, linenum, from);
    return 0;
  }

  CheckPadding(ptr);


  MemHash->Table[i].Data.FreeFile = file;
  MemHash->Table[i].Data.FreeLine = linenum;

  free(MemHash->Table[i].Data.Ptr);

  MemHash->Table[i].Data.Active = 0;

  if(MemHash->Table[i].Data.Name)
    MemHash->Table[i].Data.Name = 0;

  MemHash->Table[i].Data.Protected = 0;

  MHRemove((unsigned int)MemHash->Table[i].Data.Ptr);

  return MemHash->Table[i].Data.Size;
}

/*******************************************************************************
   Function: sRenameAlloc()

Description: Changes the identification name of the pointer.

     Inputs: *ptr  - Pointer to change the name of
             *name - Pointer to a new string containg the new name for the ptr

    Outputs: 1 - Name updated
             0 - The supplied pointer could not be found.
*******************************************************************************/
int sRenameAlloc(void *ptr, char *name, char *file, int linenum)
{
  int i = GetAllocIndex(ptr);

  if(i == -1)
    return 0;

  MemHash->Table[i].Data.Name = name;
  MemHash->Table[i].Data.File = file;
  MemHash->Table[i].Data.Line = linenum;

  return 1;
}

/*******************************************************************************
   Function: ProtectAlloc()

Description: "Protects" a chunk of allocated memory. A protected chuck of memory
                will not be freed by GarbageCollect, and not printed out by
                DumpAllocatedMemNonProtected.

     Inputs: *ptr - Pointer to add the protection flag to.

    Outputs: N/A
*******************************************************************************/
void ProtectAlloc(void *ptr)
{
  int i = GetAllocIndex(ptr);

  if(i == -1)
    return;

  MemHash->Table[i].Data.Protected = 1;

  return;
}

/*******************************************************************************
   Function: UnProtectAlloc()

Description: Removes protectio from a pointer. A unprotected pointer will be
              freed by GarbageCollect, and printed out by all the DumpAllocMem
              functions.

     Inputs: *ptr - Pointer to unprotect

    Outputs: N/A
*******************************************************************************/
void UnProtectAlloc(void *ptr)
{
  int i = GetAllocIndex(ptr);

  if(i == -1)
    return;

  MemHash->Table[i].Data.Protected = 0;

  return;
}

/*******************************************************************************
   Function: ProtectAlloc()

Description: "Protects" all of the allocated memory. A protected chuck of memory
                will not be freed by GarbageCollect, and not printed out by
                DumpAllocatedMemNonProtected.

     Inputs: *ptr - Pointer to add the protection flag to.

    Outputs: N/A
*******************************************************************************/
void ProtectAll(void)
{
  unsigned int i;

  if(!MemHash) return;

  for(i = 0; i < MemHash->TableSize; ++i)
    if(MemHash->Table[i].State == OCCUPIED && MemHash->Table[i].Data.Active)
        MemHash->Table[i].Data.Protected = 1;

  return;
}

/*******************************************************************************
   Function: GarbageCollect()

Description: Frees the memory being used by all non protected variables.

     Inputs: N/A

    Outputs: Total number of bytes freed.
*******************************************************************************/
int GarbageCollect(void)
{
  unsigned int i;
  int size = 0;

  printf("Garbage Collect Begining...\n<------------------->\n");


  for(i = 0; i < MemHash->TableSize; ++i)
  {
    if(MemHash->Table[i].State == OCCUPIED && MemHash->Table[i].Data.Active)
      if(!MemHash->Table[i].Data.Protected)
      {
        printf("Garabage Collect - Removed: %p\n", MemHash->Table[i].Data.Ptr);
        printf("Name: %s\n", MemHash->Table[i].Data.Name);
        PrintSize(MemHash->Table[i].Data.Size);
        printf("\n");
        size += MemHash->Table[i].Data.Size;
        sfree((unsigned char *)MemHash->Table[i].Data.Ptr + PADDING_BEFORE, "Garbage Collection", 0, FROM_MALLOC);
      }
  }

  printf("\nFreed ");
  PrintSize(size);
  printf("<------------------->\n\n");

  return size;
}


void sBreakOnError(int state)
{
  #if DEBUG_WINDOWS
    BreakOnError = state;
  #else
    state = state;
  #endif
}

/*******************************************************************************
   Function: DumpAllocStatus()

Description: Prints out general information about the allocated memory
                <--------------->
                Active Pointers: Total Number of Pointers that are in use.
                Active Bytes   : Total Size of all the allocated memory
                ----
                Protected      : Number of Protected pointers
                Protected Size : Size of all the Protected pointers
                ---
                Non Protected  : Number of NonProtected (temporary) pointers
                Non Protected S: Size of all the NonProtected poitners
                ---
                Padding Size   : Additional number of bytes alloc for each ptr
                Total Alloc    : Total size of all pointers, plus the padding
                <--------------->

     Inputs: N/A

    Outputs: N/A - Current alloc status printed to the command prompt
*******************************************************************************/
void DumpAllocStatus(void)
{
  unsigned int i;
  int num    = 0; /* Total Number of pointers                                 */
  int size   = 0; /* Size of all the pointers                                 */
  int pnum   = 0; /* Number of All the Protected Pointers                     */
  int psize  = 0; /* Size of all the Protected Pointers                       */
  int npnum  = 0; /* Number of All the NonProtected Pointers                  */
  int npsize = 0; /* Size of all the NonProtected Pointers                    */

  if(!MemHash) return;

  printf("Current Allocation Status\n<--------------->\n");


  for(i = 0; i < MemHash->TableSize; ++i)
  {
    if(MemHash->Table[i].State == OCCUPIED && MemHash->Table[i].Data.Active && MemHash->Table[i].Data.Line)
    {
      ++num;
      size += MemHash->Table[i].Data.Size;

      if(MemHash->Table[i].Data.Protected)
      {
        ++pnum;
        psize += MemHash->Table[i].Data.Size;
      }
      else
      {
        ++npnum;
        npsize += MemHash->Table[i].Data.Size;
      }
    }
  }

  printf("Active Pointers: %d\n", num   );
  printf("Active Bytes   : "            ); PrintSize(size);
  printf("----\n"                       );
  printf("Protected      : %d\n", pnum);
  printf("Protected Size : "); PrintSize(psize);
  printf("----\n"                       );
  printf("Non Protected  : %d\n", npnum);
  printf("Non Protected S: "); PrintSize(npsize);
  printf("----\n"                       );
  printf("Padding Size   : "            ); PrintSize(num * (PADDING_BEFORE + PADDING_AFTER));
  printf("Total Alloc    : "            ); PrintSize(num * (PADDING_BEFORE + PADDING_AFTER) + size);
  printf("<--------------->\n\n"        );
}

/*******************************************************************************
   Function: DumpAllocatedMem()

Description: Prints out a detailed list of all active pointers.
                  0x00000000   <- Pointer
                  File: asdf   <- File Name Data was allocated int
                  Line: 0      <- Line in the file the Data was allocated
                  Size: 0      <- Number of bytes allocated to the pointer
                  Name: asdf   <- Name of the pointer (if it has one)
                  String Dump  <- Attempt to print out the ptr as a string

     Inputs: N/A

    Outputs: N/A - Pointer data is printed to the command prompt
*******************************************************************************/
void DumpAllocatedMem(void)
{
  unsigned int i;

  if(!MemHash) return;

  printf("\nDumping Alloc Data\n");
  printf("------------------\n");


  for(i = 0; i < MemHash->TableSize; ++i)
    if(MemHash->Table[i].State == OCCUPIED && MemHash->Table[i].Data.Active == 1 && MemHash->Table[i].Data.Line)
  {
    printf("%p\nFile: %s\nLine: %d\nSize: ", (unsigned char *)MemHash->Table[i].Data.Ptr + PADDING_BEFORE, MemHash->Table[i].Data.File, MemHash->Table[i].Data.Line);
    PrintSize(MemHash->Table[i].Data.Size);

    if(MemHash->Table[i].Data.Name)
      printf("Name: %s\n", MemHash->Table[i].Data.Name);

    #if DUMPSTRING == 1
      printf("String Dump: |%s|", (char *)MemHash->Table[i].Data.Ptr + PADDING_BEFORE);
    #endif

    printf("\n\n");
  }

  printf("------------------\n\n");
}

/*******************************************************************************
   Function: DumpAllocatedMemMin()

Description: Dumps a compressed list of pointers to the command prompt.
                  0x00000000  adsf:00000: xxxxxx  |xxxxxxxxxxxxxxxxx|
                  <Pointer> <File>:<Line> <Name>  <Printed as string>

     Inputs: N/A

    Outputs: N/A - Pointer data is printed to the command prompt
*******************************************************************************/
void DumpAllocatedMemMin(void)
{
  unsigned int i;

  if(!MemHash) return;

  printf("\nDumping Alloc Data\n");
  printf("------------------\n");


  for(i = 0; i < MemHash->TableSize; ++i)
    if(MemHash->Table[i].State == OCCUPIED && MemHash->Table[i].Data.Active == 1 && MemHash->Table[i].Data.Line)
  {
    printf("%p %35s:%5d: ", (unsigned char *)MemHash->Table[i].Data.Ptr + PADDING_BEFORE,MemHash->Table[i].Data.File,MemHash->Table[i].Data.Line);

    if(MemHash->Table[i].Data.Name)
      printf("%s ", MemHash->Table[i].Data.Name);

    #if DUMPSTRING == 1
      printf("|%s|\n", (char *) MemHash->Table[i].Data.Ptr + PADDING_BEFORE);
    #else
      printf("\n");
    #endif

  }

  printf("------------------\n\n");
}

/*******************************************************************************
   Function: DumpAllocatedMemNonProtected()

Description: Dumps a compressed list of pointers to the command prompt.
              Protected Pointers are NOT printed out. Only non protected
              pointers are printed.
                  0x00000000  adsf:00000: xxxxxx  |xxxxxxxxxxxxxxxxx|
                  <Pointer> <File>:<Line> <Name>  <Printed as string>

     Inputs: N/A

    Outputs: N/A - Pointer data is printed to the command prompt
*******************************************************************************/
void DumpAllocatedMemNonProtected(void)
{
  unsigned int i;
  int size = 0;

  if(!MemHash) return;

  printf("\nDumping Nonprotected Mem\n");
  printf("------------------\n");


  for(i = 0; i < MemHash->TableSize; ++i)
    if(MemHash->Table[i].State == OCCUPIED && MemHash->Table[i].Data.Active == 1 && MemHash->Table[i].Data.Line)
  {
    if(MemHash->Table[i].Data.Protected)
      continue;

    size += MemHash->Table[i].Data.Size;

    printf("%p %35s:%5d: ", (unsigned char *)MemHash->Table[i].Data.Ptr + PADDING_BEFORE,MemHash->Table[i].Data.File,MemHash->Table[i].Data.Line);

    if(MemHash->Table[i].Data.Name)
      printf("%s ", MemHash->Table[i].Data.Name);

    #if DUMPSTRING == 1
      printf("|%s|\n", (char *) MemHash->Table[i].Data.Ptr + PADDING_BEFORE);
    #else
      printf("\n");
    #endif

  }

  printf("\nWasted ");
  PrintSize(size);
  printf("------------------\n\n");
}

/*******************************************************************************
   Function: CheckAllPadding()

Description: Iterates through all of the active pointers, and checks their
               padding.

     Inputs: N/A

    Outputs: N/A
*******************************************************************************/
void CheckAllPadding(void)
{
  unsigned int i;

  if(!MemHash) return;

  for(i = 0; i < MemHash->TableSize; ++i)
    if(MemHash->Table[i].State == OCCUPIED && MemHash->Table[i].Data.Active)
      CheckPadding((unsigned char *)MemHash->Table[i].Data.Ptr + PADDING_BEFORE);

  return;
}


unsigned int sIsABasePointer(void *ptr)
{
  int i = GetAllocIndex(ptr);

  if(i == -1)
    return 0;

  if(MemHash->Table[i].State == OCCUPIED)
    return 1;
  return 0;
}

unsigned int sIsAValidPointer(void *ptr)
{
  unsigned int i;

  if(sIsABasePointer(ptr))
    return 1;

  for(i = 0; i < MemHash->TableSize; ++i)
    if(MemHash->Table[i].State == OCCUPIED && MemHash->Table[i].Data.Active)
    {
      Alloc_str &data  = MemHash->Table[i].Data;
      void      *start = (char *)data.Ptr + PADDING_BEFORE;
      void      *end   = (char *)data.Ptr + data.Size + PADDING_BEFORE;

      if(start <= ptr && ptr < end)
        return 1;
    }

  return 0;
}

unsigned int sPointerSize(void *ptr)
{
  int i = GetAllocIndex(ptr);

  if(i == -1)
    return 0;

  if(MemHash->Table[i].State == OCCUPIED)
    return MemHash->Table[i].Data.Size;
  return 0;
}

void PrintFromMiss(int i, char *file, int linenum, int from)
{
  printf("ERROR: Mismatch ");
  switch(MemHash->Table[i].Data.Flags)
  {
    case FROM_MALLOC:  printf("malloc() ");  break;
    case FROM_NEW:     printf("new ");       break;
    case FROM_NEW_A:   printf("new[] ");     break;
  }
  printf("with ");
  switch(from)
  {
    case FROM_MALLOC:  printf("free() ");    break;
    case FROM_NEW:     printf("delete ");    break;
    case FROM_NEW_A:   printf("delete[] ");  break;
    case FROM_REALLOC: printf("realloc() "); break;
  }
  printf("\n");
  if(file)
  {
    printf("       Generated at:\n");
    printf("         File: %s\n", file);
    printf("         Line: %d\n", linenum);
  }
  printf("       Originally Allocated at:\n");
  printf("         ");
  PrintPointerData(i);

  #if DEBUG_WINDOWS
      /* You arrived here if you missmatched a new with a free(). Check       *
       *   the console or output file for more details.                       */
    if(BreakOnError) __asm int 3;
  #endif
}

/*******************************************************************************
   Function: GetAllocIndex()

Description: Retrieves the Index of the specifed pointer.

     Inputs: *ptr - Pointer to get the index of.

    Outputs: 0+ - Index of the pointer.
             -1 - Supplied pointer is not valid, or not created by this module
*******************************************************************************/
int GetAllocIndex(void *ptr)
{
  if(ptr == 0)
    return -1;

  return MHFind((unsigned long)((unsigned char *)ptr - PADDING_BEFORE));
}

/*******************************************************************************
   Function: CheckOldRecords()

Description: Retrieves the Index of a pointer that was already freed, and still
             in the Alloc_str

     Inputs: *ptr - Pointer to get the index of.

    Outputs: 0+ - Index of the pointer.
             -1 - Supplied pointer is not valid, or not created by this module
*******************************************************************************/
int CheckOldRecords(void *ptr)
{
  if(ptr == 0)
    return -1;

  return MHFindDeleted((unsigned long)((unsigned char *)ptr - PADDING_BEFORE));
}

/*******************************************************************************
   Function: PrintSize()

Description: Prints out the size, in a nice and compressed state, followed by
                a new line.
                    1.23 GB, 1.23 MB, 1.23 KB, or 123 Bytes

     Inputs: size - Size to format and print to the command prompt.

    Outputs: N/A - Size is printed to the command prompt, followed by a new line
*******************************************************************************/
void PrintSize(int size)
{
  int gb;
  int mb;
  int kb;
  int  b;

  gb = size / 1073741824;
  mb = (size % 1073741824) / 1048576;
  kb = (size % 1048576) / 1024;
  b  = size % 1024;

  if(gb)
    printf("%d.%d GB\n", gb, (int)(mb /1024.0 * 100));
  else
  if(mb)
    printf("%d.%d MB\n", mb, (int)(kb /1024.0 * 100));
  else
  if(kb)
    printf("%d.%d KB\n", kb, (int)(b /1024.0 * 100));
  else
    printf("%d Bytes\n", b);
}

/*******************************************************************************
   Function: MarkPadding()

Description: Writes the specified signature (PADDINGBYTE) to the padding
               attached to each block of memory.

     Inputs: *ptr - Poitner to the block of memory. This should be the pointer
                      returned from malloc().
             size - Size of the requested memory. IE, this does not include
                      the padding attached to the memory.

    Outputs: N/A
*******************************************************************************/
void MarkPadding(void *ptr, int size)
{
  int i;
  unsigned char *ucptr = (unsigned char *)ptr;

  for(i = 0; i < PADDING_BEFORE; i++)
    ucptr[i] = PADDINGBYTE;

  for(i = 0; i < PADDING_AFTER; i++)
    ucptr[size + PADDING_BEFORE + i] = PADDINGBYTE;
}

/*******************************************************************************
   Function: CheckPadding()

Description: Checks the padding bytes of the specfied block of memory. If the
               signatures do not match, error messages will be printed to
               stdout.

     Inputs: ptr - Pointer to check.

    Outputs: N/A - Messages may be printed to stdout
*******************************************************************************/
void CheckPadding(void *ptr)
{
  int index = GetAllocIndex(ptr);
  int i;
  unsigned char *aptr;
  int size;
  int errorcount = 0;
  int error = 0;

  if(index == -1)
    return;

  aptr = (unsigned char *)MemHash->Table[index].Data.Ptr;
  size = MemHash->Table[index].Data.Size;

  for(i = 0; i < PADDING_BEFORE; ++i)
    if(aptr[i] != PADDINGBYTE)
    {
      if(!errorcount)
      {
        printf("ERROR: Buffer underflow detected in the pointer\n");
        PrintPointerData(index);
      }

      ++errorcount;
    }

  if(errorcount > 1)
  {
    printf("ERROR: Supressed %d bytes of underflow errors\n", errorcount - 1);

    if(errorcount == PADDING_BEFORE)
      printf("ERROR: Underflow may have exceeded the padding\n\n");
    else
      printf("\n");
  }

  if(errorcount)
    error = 1;
  errorcount = 0;

  i += size;

  for(; i < size + PADDING_BEFORE + PADDING_AFTER; ++i)
    if(aptr[i] != PADDINGBYTE)
    {
      if(!errorcount)
      {
        printf("ERROR: Buffer overflow detected in the pointer\n");
        PrintPointerData(index);
      }

      ++errorcount;
    }

  if(errorcount > 1)
  {
    printf("ERROR: Supressed %d bytes of overflow errors\n", errorcount - 1);

    if(errorcount == PADDING_AFTER)
      printf("ERROR: Overflow may have exceeded the padding\n\n");
    else
      printf("\n");
  }

  #if DEBUG_WINDOWS
      /* You arrived here if you overwrote a fence. Check the console or outpu*
       *   file for more details                                              */
    if(errorcount || error) if(BreakOnError) __asm int 3;
  #endif
}

/*******************************************************************************
   Function: PrintPointerData()

Description: Prints out the pointer information in a streamline format.
               <prt> <File>: <line>: <Symbolic name> |<string dump>|

     Inputs: i - Index of the pointer to print.

    Outputs: N/A
*******************************************************************************/
void PrintPointerData(int i)
{
  printf("%p %s:%5d: ", (unsigned char *)MemHash->Table[i].Data.Ptr + PADDING_BEFORE,MemHash->Table[i].Data.File,MemHash->Table[i].Data.Line);

  if(MemHash->Table[i].Data.Name)
    printf("%s ", MemHash->Table[i].Data.Name);

  #if DUMPSTRING == 1
    printf("|%s|\n", (char *) MemHash->Table[i].Data.Ptr + PADDING_BEFORE);
  #else
    printf("\n");
  #endif

  printf("\n");
}










static void       MHInit        (void);
static void       MHGrowTable   (void);
static void       MHInitTable   (void);
static int        MHGetFreeAlloc(unsigned long Key);
static unsigned   MHHashNum     (unsigned long Key);




int MHInsert(unsigned long Key)
{
  int i;

  if(!MemHash) MHInit();
  ++MemHash->Elements;
  MemHash->Load = (float)MemHash->Elements / MemHash->TableSize;
  if(MemHash->Load > MemHash->MaxLoad)
  {
    MHGrowTable();
    ++MemHash->Elements;
  }

  i = MHGetFreeAlloc(Key);
  MemHash->Table[i].Key   = Key;
  MemHash->Table[i].State = OCCUPIED;

  return i;
}

void MHRemove(unsigned long Key)
{
  unsigned i;

  if(!MemHash) MHInit();
  i = MHFind(Key);

  //MemHash->Table[i].Key = 0;
  --MemHash->Elements;

  MemHash->Table[i].State = DELETED;
}


int MHFind(unsigned long Key)
{
  unsigned i;
  unsigned p = 0; /* Number of probes in this loop. Prevents infinate loop    */

  if(!MemHash) MHInit();
  i = MHHashNum(Key);

  while(MemHash->Table[i].State != UNOCCUPIED && ++p < MemHash->TableSize)
  {
    if(MemHash->Table[i].Key == Key && MemHash->Table[i].State == OCCUPIED)
      return i;

    if(++i >= MemHash->TableSize)
      i -= MemHash->TableSize;
  }

  return -1;
}


int MHFindDeleted(unsigned long Key)
{
  unsigned i;
  unsigned p = 0; /* Number of probes in this loop. Prevents infinate loop    */

  if(!MemHash) MHInit();
  i = MHHashNum(Key);

  while(MemHash->Table[i].State != UNOCCUPIED && ++p < MemHash->TableSize)
  {
    if(MemHash->Table[i].Key == Key && MemHash->Table[i].State == DELETED)
      return i;

    if(++i >= MemHash->TableSize)
      i -= MemHash->TableSize;
  }

  return -1;
}

void MHInit(void)
{
  MemHash            = (struct MemHash_str *)malloc(sizeof(struct MemHash_str));
  MemHash->TableSize = ALLOCSIZE;
  MemHash->Table     = (struct MHSlot_str *)malloc(sizeof(struct MHSlot_str) * MemHash->TableSize);
  MemHash->Elements  = 0;
  MemHash->Load      = 0.0;
  MemHash->MaxLoad   = 0.8f;
  MemHash->Growth    = 1.5f;

  MHInitTable();
}

void MHGrowTable(void)
{
  int i;
  int j;
  struct MHSlot_str *table = MemHash->Table;
  int tsize = MemHash->TableSize;
  MemHash->TableSize = (unsigned)(MemHash->TableSize * MemHash->Growth);
  MemHash->Table = (struct MHSlot_str *)malloc(sizeof(struct MHSlot_str) * MemHash->TableSize);

  MHInitTable();

  if(table)
  {
    for(i = 0; i < tsize; ++i)
      if(table[i].State == OCCUPIED)
      {
        j = MHInsert(table[i].Key);
        if(j != -1)
          MemHash->Table[j].Data = table[i].Data;
      }
      else
      if(table[i].State == DELETED)
      {
        j = MHInsert(table[i].Key);
        if(j != -1)
          MemHash->Table[j].Data = table[i].Data;
        MHRemove(table[i].Key);
      }
  }

  free(table);
}


void MHInitTable(void)
{
  unsigned int i;
  for( i = 0; i < MemHash->TableSize; ++i)
    MemHash->Table[i].State = UNOCCUPIED;

  MemHash->Elements  = 0;
  MemHash->Load      = 0.0;
}


int MHGetFreeAlloc(unsigned long Key)
{
  unsigned i;
  unsigned p = 0; /* Number of probes in this loop. Prevents infinate loop    */

  if(!MemHash) MHInit();
  i = MHHashNum(Key);

  while(++p < MemHash->TableSize)
  {
    if(MemHash->Table[i].State == DELETED || MemHash->Table[i].State == UNOCCUPIED)
      return i;

    if(++i >= MemHash->TableSize)
      i -= MemHash->TableSize;
  }

  return -1;
}



unsigned MHHashNum(unsigned long key)
{
   key = (key + 0x0008EF18) + (key <<  5);
   key = (key ^ 0x000CE236) ^ (key >> 16);
   key = (key + 0x0002A45D) + (key <<  3);
   key = (key ^ 0x00030CDA) ^ (key << 14);
   key = (key + 0x00037823) + (key <<  9);
   key = (key ^ 0x0002CB17) ^ (key >> 12);
   return key % MemHash->TableSize;

}


#ifdef __cplusplus
#if SAFEALLOC_OVERLOAD_CPP_ALLOC

#if !DEBUG_WINDOWS
  #include <new>

  void *operator new(unsigned int size) throw(std::bad_alloc)
  {
    void *ret;

    if(!size)
      size = sizeof(int);

    ret = snmalloc(size, 0, 0, 0, FROM_NEW);

    if(!ret)
      throw(std::bad_alloc());

    return ret;
  }

  void *operator new(unsigned int size, const std::nothrow_t&) throw()
  {
    return snmalloc(!size ? sizeof(int) : size, 0, 0, 0, FROM_NEW);
  }

  void *operator new[](unsigned int size) throw(std::bad_alloc)
  {
    void *ret;

    if(!size)
      size = sizeof(int);

    ret = snmalloc(size, 0, 0, 0, FROM_NEW_A);

    if(!ret)
      throw(std::bad_alloc());

    return ret;
  }

  void *operator new[](unsigned int size, const std::nothrow_t&) throw()
  {
    return snmalloc(!size ? sizeof(int) : size, 0, 0, 0, FROM_NEW_A);
  }

  void operator delete(void *ptr) throw()
  {
    sfree(ptr, 0, 0, FROM_DEL);
  }

  void  operator delete(void * ptr, const std::nothrow_t&) throw()
  {
    sfree(ptr, 0, 0, FROM_DEL);
  }

  void  operator delete[](void * ptr) throw()
  {
    sfree(ptr, 0, 0, FROM_DEL_A);
  }

  void  operator delete[](void * ptr, const std::nothrow_t&) throw()
  {
    sfree(ptr, 0, 0, FROM_DEL_A);
  }
#else
  #include <windows.h>
  #include <dbghelp.h>
  #include <intrin.h>
  #include <new>

  #pragma comment( lib, "dbghelp" )
  static int  GetAddrMap      (DWORD64 addr);

  void *operator new(unsigned int size) throw(std::bad_alloc)
  {
    int i = GetAddrMap((DWORD64)_ReturnAddress());
    void *ret;

    if(!size)
      size = sizeof(int);

    if(i == -1)
      ret = snmalloc(size, 0, 0, 0, FROM_NEW);
    else
      ret = snmalloc(size, 0, AddrHash->Table[i].Data.File, AddrHash->Table[i].Data.Line, FROM_NEW);

    if(!ret)
      throw(std::bad_alloc());

    return ret;
  }

  void *operator new(unsigned int size, const std::nothrow_t&) throw()
  {
    int i = GetAddrMap((DWORD64)_ReturnAddress());
    void *ret;

    if(!size)
      size = sizeof(int);

    if(i == -1)
      ret = snmalloc(size, 0, 0, 0, FROM_NEW);
    else
      ret = snmalloc(size, 0, AddrHash->Table[i].Data.File, AddrHash->Table[i].Data.Line, FROM_NEW);

    return ret;
  }

  void *operator new[](unsigned int size) throw(std::bad_alloc)
  {
    int i = GetAddrMap((DWORD64)_ReturnAddress());
    void *ret;

    if(!size)
      size = sizeof(int);

    if(i == -1)
      ret = snmalloc(size, 0, 0, 0, FROM_NEW_A);
    else
      ret = snmalloc(size, 0, AddrHash->Table[i].Data.File, AddrHash->Table[i].Data.Line, FROM_NEW_A);

    if(!ret)
      throw(std::bad_alloc());

    return ret;
  }

  void *operator new[](unsigned int size, const std::nothrow_t&) throw()
  {
    int i = GetAddrMap((DWORD64)_ReturnAddress());
    void *ret;

    if(!size)
      size = sizeof(int);

    if(i == -1)
      ret = snmalloc(size, 0, 0, 0, FROM_NEW_A);
    else
      ret = snmalloc(size, 0, AddrHash->Table[i].Data.File, AddrHash->Table[i].Data.Line, FROM_NEW_A);

    return ret;
  }

  void operator delete(void *ptr) throw()
  {
    int i = GetAddrMap((DWORD64)_ReturnAddress());

    if(!ptr)
      return;

    if(i == -1)
      sfree(ptr, 0, 0, FROM_DEL);
    else
      sfree(ptr, AddrHash->Table[i].Data.File, AddrHash->Table[i].Data.Line, FROM_DEL);
  }

  void  operator delete(void * ptr, const std::nothrow_t&) throw()
  {
    int i = GetAddrMap((DWORD64)_ReturnAddress());

    if(!ptr)
      return;

    if(i == -1)
      sfree(ptr, 0, 0, FROM_DEL);
    else
      sfree(ptr, AddrHash->Table[i].Data.File, AddrHash->Table[i].Data.Line, FROM_DEL);
  }

  void  operator delete[](void * ptr) throw()
  {
    int i = GetAddrMap((DWORD64)_ReturnAddress());

    if(!ptr)
      return;

    if(i == -1)
      sfree(ptr, 0, 0, FROM_DEL_A);
    else
      sfree(ptr, AddrHash->Table[i].Data.File, AddrHash->Table[i].Data.Line, FROM_DEL_A);
  }

  void  operator delete[](void * ptr, const std::nothrow_t&) throw()
  {
    int i = GetAddrMap((DWORD64)_ReturnAddress());

    if(!ptr)
      return;

    if(i == -1)
      sfree(ptr, 0, 0, FROM_DEL_A);
    else
      sfree(ptr, AddrHash->Table[i].Data.File, AddrHash->Table[i].Data.Line, FROM_DEL_A);
  }



  static void GetAddrMapName(DWORD64 addr, int i);
  static void GetAddrMapLine(DWORD64 addr, int i);

  int GetAddrMap(DWORD64 addr)
  {
    struct MemHash_str *hash = MemHash;
    int i;

    MemHash = AddrHash;

    if(!AddrHash)
    {
        /* Load the symbols and line numbers */
      SymInitialize(GetCurrentProcess(), 0, 1);
      SymSetOptions(SymGetOptions() | SYMOPT_LOAD_LINES);
    }

    i = MHFind((unsigned long)addr);

    if(i == -1)
      if((i = MHInsert((unsigned long)addr)) != -1)
      {
        GetAddrMapName(addr, i);
        GetAddrMapLine(addr, i);
      }

    AddrHash = MemHash;
    MemHash  = hash;

    return i;
  }


  void GetAddrMapName(DWORD64 addr, int i)
  {
    char   *name = "";
    IMAGEHLP_LINE line;
    unsigned long off = 0;

    line.SizeOfStruct = sizeof(line);
    if(SymGetLineFromAddr(GetCurrentProcess(), (DWORD)addr, &off, &line))
      name = line.FileName;

    MemHash->Table[i].Data.File = (char *)malloc((unsigned int)strlen(name) + 1);
    memcpy(MemHash->Table[i].Data.File, name, strlen(name) + 1);
  }

  void GetAddrMapLine(DWORD64 addr, int i)
  {
    IMAGEHLP_LINE line;
    unsigned long off = 0;

    line.SizeOfStruct = sizeof(line);
    if(SymGetLineFromAddr(GetCurrentProcess(), (DWORD)addr, &off, &line))
      MemHash->Table[i].Data.Line = (int)line.LineNumber;
    else
      MemHash->Table[i].Data.Line = 0;
  }
#endif /* DEBUG_WINDOWS */

#endif /* SAFEALLOC_OVERLOAD_CPP_ALLOC */
#endif /* __cplusplus */
#endif /* SAFEALLOC_ENABLED */
