/*******************************************************************************
filename      SafeAlloc.h
author        Craig Williams
e-mail        cwilliam@digipen.edu

Brief Description:
  Notice: SafeAlloc.h must be included in every .c/cpp file. If you do not,
            and the memory debugger is enabled, the heap may become corrupted
            since calls to *alloc() and free() will not be intercepted in the
            files that do not include SafeAlloc.h.

  Provides various memory debugging functions. Including this header will
    overwrite *alloc(), free(), new and delete. When you allocate some memory,
    a specified number of padding bytes will be added to each end of the memory.
    These padding blocks will have a signature written to them. If the signature
    changes, you will be notifed.

  Functions:
    s*alloc()  - Overloaded version of the specified alloc()
    sn*alloc() - Same as above; however, you can attach a meaningful name to the
                   memory.
    sfree()    - Overloaded version of free()

  Memory Profiling Functions:
    sRenameAlloc()   - Allows you to attach a meaningful name to a pointer. This
                         name will be printed out with the pointer when ever you
                         call a Dump*().
    ProtectAlloc()   - Marks a pointer as "protected". This prevents
                         GarbageCollect() from freeing the data when called.
                         This is also used to signify that the specified pointer
                         is supposed to be there.
    UnProtectAlloc() - Removes the "protected" flag.
    ProtectAll()     - Sets all the active pointers as protected.
    GarbageCollect() - Removes all memory that was not marked as protected. This
                         function should not be used.

  General Debug Functions:
    sBreakOnError()       - Should SafeAlloc() break on a error?
    sIsABasePointer()     - Returns 1 if the pointer is a valid base pointer
    sPointerSize()        - If the pointer is a base pointer, size of the
                              allocated memory. Otherwise, 0.

  Memory Status Functions:
    DumpAllocStatus()     - Prints out how much memory is being used, as well
                              as how many pointers there are. Calling this
                              function at the end of main() or similar will
                              give you a report of the amount of memory was
                              leaked.
    DumpAllocatedMem()    - Lists all the pointers that the memory debugger is
                              currently tracking including the alloc location,
                              size, and a symbol name if it has one.
    DumpAllocatedMemMin() - Same as above; however, the output is far more
                              streamlined. IE, 1-2 lines per pointer.
    DumpAllocatedMemNonProtected() - Same as DumpAllocatedMem(), but it only
                                       dumps non protected pointers.
    CheckAllPadding()     - Checks every pointer for over/under writes. This is
                              done by checking the padding area attached to each
                              pointer.
*******************************************************************************/
#ifndef SAFEALLOC_H
#define SAFEALLOC_H

#ifndef SAFEALLOC_ENABLED
  #define SAFEALLOC_ENABLED            1
#endif
#ifndef SAFEALLOC_OVERLOAD_C_ALLOC
  #define SAFEALLOC_OVERLOAD_C_ALLOC   1
#endif
#ifndef SAFEALLOC_OVERLOAD_CPP_ALLOC
  #define SAFEALLOC_OVERLOAD_CPP_ALLOC 1
#endif

#if SAFEALLOC_ENABLED
#include <stdlib.h>

#if SAFEALLOC_OVERLOAD_C_ALLOC
  #define    malloc(size           )   smalloc(size,           __FILE__, __LINE__)
  #define   nmalloc(size, name     )  snmalloc(size, name,     __FILE__, __LINE__, 0)
  #define    calloc(ele, size      )   scalloc(ele, size,      __FILE__, __LINE__)
  #define   ncalloc(ele, size, name)  sncalloc(ele, size, name,__FILE__, __LINE__)
  #define   realloc(ptr, size      )  srealloc(ptr, size,      __FILE__, __LINE__)
  #define  nrealloc(ptr, size, name) snrealloc(ptr, size, name,__FILE__, __LINE__)
  #define      free(ptr            )     sfree(ptr,            __FILE__, __LINE__, 0)
#endif

#define RenameAlloc(ptr, name   ) sRenameAlloc(ptr, name,    __FILE__, __LINE__)

void *smalloc   (unsigned int size,                                    char *file, int linenum);
void *snmalloc  (unsigned int size,  char          *name,              char *file, int linenum, int from);
void *srealloc  (void        *ptr,   unsigned int   size,              char *file, int linenum);
void *snrealloc (void        *ptr,   unsigned int   size, char *name,  char *file, int linenum);
void *scalloc   (unsigned int ele,   unsigned int   numofele,          char *file, int linenum);
void *sncalloc  (unsigned int esize, unsigned int   noe,  char *name,  char *file, int linenum);
int   sfree     (void        *ptr,                                     char *file, int linenum, int from);
int sRenameAlloc(void        *ptr,   char          *name,              char *file, int linenum);

  /* Basic Profiling */
void ProtectAlloc  (void *ptr);
void UnProtectAlloc(void *ptr);
void ProtectAll    (void     );
int  GarbageCollect(void     );

  /* Configuration */
void sBreakOnError(int state);

  /* Debug Ouput */
void DumpAllocStatus             (void);
void DumpAllocatedMem            (void);
void DumpAllocatedMemMin         (void);
void DumpAllocatedMemNonProtected(void);
void CheckAllPadding             (void);

unsigned int sIsABasePointer     (void *mem);
unsigned int sIsAValidPointer    (void *mem); /* Base pointer or in valid rang*/
unsigned int sPointerSize        (void *mem); /* Must be a base pointer       */

#ifdef __cplusplus
  #if SAFEALLOC_OVERLOAD_CPP_ALLOC
    #include <new>
      /* Disable the warning that the compiler ignores the throw type         */
    #if DEBUG_WINDOWS
      #pragma warning(disable : 4290)
    #endif

    void *operator new     (unsigned int size) throw(std::bad_alloc);
    void *operator new     (unsigned int size, const std::nothrow_t&) throw();
    void *operator new[]   (unsigned int size) throw(std::bad_alloc);
    void *operator new[]   (unsigned int size, const std::nothrow_t&) throw();
    void  operator delete  (void * ptr)                               throw();
    void  operator delete  (void * ptr, const std::nothrow_t&)        throw();
    void  operator delete[](void * ptr)                               throw();
    void  operator delete[](void * ptr, const std::nothrow_t&)        throw();
  #endif /* SAFEALLOC_OVERLOAD_CPP_ALLOC */
#endif   /* __cplusplus            */
#endif /* SAFEALLOC_ENABLED */
#endif /* SAFEALLOC_H */
