#include <stdio.h>

#include "../OS Specific/DirList.h"
#include "../OS Specific/DirString.h"
#include "../OS Specific/SystemCalls.h"
#include "../FileIO/Parser.h"
#include "../ConfigFile.h"
#include "DirectoryTree.h"
#include "FilePaths.h"
#include "Management.h"
#include "File.h"

#include "Directory.h"

#include "../Debug/Debug.h"

CString FindMakeFile(const CString &dir);

Directory::Directory(const CString &path, DirectoryTree *tree, ConfigFile *cfg)
{
  DirList dir(path.c_str());

  // tree is on the stack. Don't Validate // VALIDATE_POINTER(tree)

  Tree = tree;
  if(!Tree->Root)
    Tree->Root = this;

  Config = cfg;

  Path = path;

  CheckForConfigFile();

  GetFiles(dir);
  GetSubDirectories(dir);
}

Directory::~Directory(void)
{
  unsigned int i;
  
  for(i = 0; i < Headers.size(); ++i)
    delete Headers[i];
    
  for(i = 0; i < Sources.size(); ++i)
    delete Sources[i];
    
  for(i = 0; i < Dirs.size(); ++i)
    delete Dirs[i];
}

void Directory::CheckTree(void)
{
  unsigned int i;
  
  VALIDATE_POINTER(Config)
  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Dirs.size(); ++i)
      VALIDATE_POINTER(Dirs[i])}
  #endif
  
  Config->DBase.CheckTree(this);

  for(i = 0; i < Dirs.size(); ++i)
    Dirs[i]->CheckTree();
}

void Directory::CheckIncludes(void)
{
  unsigned int i;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Headers.size(); ++i)
      VALIDATE_POINTER(Headers[i])}
    {for(unsigned int i = 0; i < Sources.size(); ++i)
      VALIDATE_POINTER(Sources[i])}
    {for(unsigned int i = 0; i < Dirs.size(); ++i)
      VALIDATE_POINTER(Dirs[i])}
  #endif

  for(i = 0; i < Headers.size(); ++i)
    Headers[i]->CheckIncludes();

  for(i = 0; i < Sources.size(); ++i)
    Sources[i]->CheckIncludes();

  for(i = 0; i < Dirs.size(); ++i)
    Dirs[i]->CheckIncludes();
}


void Directory::CheckForConfigFile(void)
{
  CString file = FindMakeFile(Path);

  // tree is on the stack. Don't Validate // VALIDATE_POINTER(Tree)

  if(!Config)
  {
    if(!file.size())
      file = Path + "build.dbbs";

    Config = new ConfigFile(file);
    Tree->Configs.push_back(Config);
  }
  else
  if(file.size())
  {
    Config = new ConfigFile(file);
    Tree->Configs.push_back(Config);
  }
  
  #if SAFEALLOC_ENABLED
    RenameAlloc(Config, "Config File");
  #endif
}

void Directory::SaveConfig(void)
{
  unsigned int i;
  
  VALIDATE_POINTER(Config)
  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Dirs.size(); ++i)
      VALIDATE_POINTER(Dirs[i])}
  #endif
  
  Config->Write(Config->F.c_str());

  for(i = 0; i < Dirs.size(); ++i)
    Dirs[i]->SaveConfig();
}

void Directory::GetSubDirectories(DirList &_dir)
{
  unsigned int i;
  unsigned int j;
  CVector<CString> dir = _dir.GetDirList();

  VALIDATE_POINTER(Config)
  
  for(i = 0; i < dir.size(); ++i)
  {
    for(j = 0; j < Config->IgnoreDirs.size(); ++j)
      if(dir[i] == Config->IgnoreDirs[j])
        break;
    if(j >= Config->IgnoreDirs.size())
      Dirs.push_back(new Directory(Path + dir[i] +'/', Tree, Config));
  }
}

void Directory::GetFiles(DirList &dir)
{
  unsigned int i;
  unsigned int j;
  unsigned int k;

  VALIDATE_POINTER(Config)
  
  for(i = 0; i < Config->HeaderTypes.size(); ++i)
  {
    CVector<CString> files = dir.GetFileList(Config->HeaderTypes[i].c_str());

    for(j = 0; j < files.size(); ++j)
    {
      for(k = 0; k < Config->IgnoreFiles.size(); ++k)
        if(Cmp(Config->IgnoreFiles[k].c_str(), files[j].c_str()))
          break;

      if(k >= Config->IgnoreFiles.size())
        Headers.push_back(new File(Path, files[j], Tree->Root, Config));
    }
  }

  for(i = 0; i < Config->SourceTypes.size(); ++i)
  {
    CVector<CString> files = dir.GetFileList(Config->SourceTypes[i].c_str());

    for(j = 0; j < files.size(); ++j)
    {
      for(k = 0; k < Config->IgnoreFiles.size(); ++k)
        if(Cmp(Config->IgnoreFiles[k].c_str(), files[j].c_str()))
          break;

      if(k >= Config->IgnoreFiles.size())
        Sources.push_back(new File(Path, files[j], Tree->Root, Config));
    }
  }
  
  #if SAFEALLOC_ENABLED
    for(unsigned int i = 0; i < Headers.size(); ++i)
      RenameAlloc(Headers[i], "Header File");
    for(unsigned int i = 0; i < Sources.size(); ++i)
      RenameAlloc(Sources[i], "Source File");
  #endif
}

void Directory::Clean(void)
{
  unsigned int i;
  
  VALIDATE_POINTER(Config)
  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Sources.size(); ++i)
      VALIDATE_POINTER(Sources[i])}
    {for(unsigned int i = 0; i < Dirs.size(); ++i)
      VALIDATE_POINTER(Dirs[i])}
  #endif
  
  for(i = 0; i < Sources.size(); ++i)
    if(FileExists(Sources[i]->GetObjFile()))
      remove(Sources[i]->GetObjFile().c_str());

  for(i = 0; i < Dirs.size(); ++i)
    Dirs[i]->Clean();

  DirList list((Config->ObjDir + Path).c_str());
  if(!list.GetFileList("*").size() && !list.GetDirList().size())
    DeleteDir((Config->ObjDir + Path).c_str());
}

void SplitFileName(const CString &in, CString &title, CString &ext)
{
  title = GetFile(in);
  ext   = GetExt(title);
  title.resize(title.size() - ext.size());
}

CString GetTitleOnly(const CString &in)
{
  CString title;
  CString ext;
  SplitFileName(in, title, ext);
  return title;
}

CString GetExtOnly(const CString &in)
{
  CString title;
  CString ext;
  SplitFileName(in, title, ext);
  return ext;
}

void Directory::GetSourceList(CVector<CString> &files)
{
  unsigned int i, j;
  CVector<CString> list;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Headers.size(); ++i)
      VALIDATE_POINTER(Headers[i])}
    {for(unsigned int i = 0; i < Sources.size(); ++i)
      VALIDATE_POINTER(Sources[i])}
    {for(unsigned int i = 0; i < Dirs.size(); ++i)
      VALIDATE_POINTER(Dirs[i])}
  #endif

  for(i = 0; i < Headers.size(); ++i)
    list.push_back(Headers[i]->Path + Headers[i]->Name);

  for(i = 0; i < Sources.size(); ++i)
    list.push_back(Sources[i]->Path + Sources[i]->Name);

  for(i = 0; i < list.size(); ++i)
  {
    CString f;
    CVector<CString> exts;
    CString title, ext;

    SplitFileName(list[i], title, ext);
    f = title;
    exts.push_back(ext);

    for(j = i + 1; j < list.size(); ++j)
    {
      SplitFileName(list[j], title, ext);
      if(Cmp(title.c_str(), f.c_str()))
      {
        exts.push_back(ext);
        list.erase(j--);
      }
    }


    f += exts[0];
    for(j = 1; j < exts.size(); ++j)
      f += ", " + exts[j];

    files.push_back(GetPath(list[i]) + f);
  }

  for(i = 0; i < Dirs.size(); ++i)
    Dirs[i]->GetSourceList(files);
}

void Directory::GetModifiedList(CVector<CString> &files)
{
  unsigned int i, j;
  CVector<CString> list;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Headers.size(); ++i)
      VALIDATE_POINTER(Headers[i])}
    {for(unsigned int i = 0; i < Sources.size(); ++i)
      VALIDATE_POINTER(Sources[i])}
    {for(unsigned int i = 0; i < Dirs.size(); ++i)
      VALIDATE_POINTER(Dirs[i])}
  #endif

  for(i = 0; i < Headers.size(); ++i)
    if(Headers[i]->WasModified)
      list.push_back(Headers[i]->Path + Headers[i]->Name);

  for(i = 0; i < Sources.size(); ++i)
    if(Sources[i]->WasModified)
      list.push_back(Sources[i]->Path + Sources[i]->Name);

  for(i = 0; i < list.size(); ++i)
  {
    CString f;
    CVector<CString> exts;
    CString title, ext;

    SplitFileName(list[i], title, ext);
    f = title;
    exts.push_back(ext);

    for(j = i + 1; j < list.size(); ++j)
    {
      SplitFileName(list[j], title, ext);
      if(Cmp(title.c_str(), f.c_str()))
      {
        exts.push_back(ext);
        list.erase(j--);
      }
    }

    f += exts[0];
    for(j = 1; j < exts.size(); ++j)
      f += ", " + exts[j];

    files.push_back(GetPath(list[i]) + f);
  }

  for(i = 0; i < Dirs.size(); ++i)
    Dirs[i]->GetModifiedList(files);
}

CString FindMakeFile(const CString &_dir)
{
  DirList dir(_dir);
  CVector<CString> files = dir.GetFileList("*.dbbs");

  if(files.size())
    return dir.GetCurrentDir() + files[0];
  return "";
}
