#include "DirectoryTree.h"
#include "Directory.h"
#include "../OS Specific/DirString.h"
#include "../FileIO/Parser.h"
#include "../ConfigFile.h"

#include "../Debug/Debug.h"

void PrintConfigName(const CString &file);

DirectoryTree::DirectoryTree(void)
{
  Root   = 0;
}

DirectoryTree::~DirectoryTree(void)
{
  unsigned int i;
  delete Root;

  for(i = 0; i < Configs.size(); ++i)
    delete Configs[i];
}

void DirectoryTree::BuildTree(const CString &root)
{
  delete Root;
  Root = new Directory(root, this);

  Root->CheckTree();
  Root->CheckIncludes();

  BuildLibLists();
}

void DirectoryTree::BuildLibLists(void)
{
  unsigned int i;
  unsigned int changed;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Configs.size(); ++i)
      VALIDATE_POINTER(Configs[i])}
  #endif

  for(i = 0; i < Configs.size(); ++i)
    Configs[i]->BuildDirectLibList();

  do
  {
    changed = 0;
    for(i = 0; i < Configs.size(); ++i)
      changed += Configs[i]->BuildIndirectLibList();
  }
  while(changed);
}

void DirectoryTree::SaveConfigs(void)
{
  VALIDATE_POINTER(Root)

  Root->SaveConfig();
}

void DirectoryTree::Compile(void)
{
  int pushed = 1;
  unsigned int i, j, k, l;
  CVector<ConfigFile *> cfg = Configs;
  CVector<ConfigFile *> order;
  CVector<CString>      errors;
  CVector<CString>      missing_comp;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {
      for(unsigned int i = 0; i < Configs.size(); ++i)
      {
        VALIDATE_POINTER(Configs[i])
        for(unsigned int j = 0; j < Configs[i]->DependsOn.size(); ++j)
        {
          VALIDATE_POINTER(Configs[i]->DependsOn[j].Config)
          CVector<DependencyLink> &links = Configs[i]->DependsOn[j].Links;
          for(unsigned int k = 0; k < links.size(); ++k)
          {
            VALIDATE_POINTER(links[k].Include)
            VALIDATE_POINTER(links[k].IncludedBy)
          }
        }
      }
    }
  #endif

  while(pushed && cfg.size())
  {
    pushed = 0;
    for(i = 0; i < cfg.size(); ++i)
    {
      for(j = 0; j < cfg[i]->DependsOn.size(); ++j)
      {
        for(k = 0; k < order.size(); ++k)
          if(cfg[i]->DependsOn[j] == order[k])
            break;
        if(k >= order.size())
          break;
      }

      if(cfg[i]->IsLibrary || j >= cfg[i]->DependsOn.size())
      {
        order.push_back(cfg[i]);
        cfg.erase(i--);
        pushed = 1;
      }
    }
  }

  if(cfg.size())
  {
    printf("\nCannot Compile - Circular Dependency Between:\n");
    for(i = 0; i < cfg.size(); ++i)
    {
      printf("  %s\n", cfg[i]->F.c_str());
      for(j = 0; j < cfg[i]->DependsOn.size(); ++j)
        for(k = 0; k < cfg.size(); ++k)
          if(cfg[i]->DependsOn[j] == cfg[k])
          {
            unsigned int len = 0;
            printf("    %s\n", cfg[k]->F.c_str());
            CVector<DependencyLink> &links = cfg[i]->DependsOn[j].Links;

            for(l = 0; l < links.size(); ++l)
            {
              CString str = links[l].IncludedBy->Path + links[l].IncludedBy->Name;
              if(str.length() > len)
                len = str.length();
            }

            for(l = 0; l < links.size(); ++l)
            {
              CString str(links[l].IncludedBy->Path + links[l].IncludedBy->Name);
              str.resize(len, ' ');
              printf("      %s  includeds  %s\n", str.c_str(),
                                                  (links[l].Include->Path    + links[l].Include->Name).c_str());
            }
            break;
          }
      printf("\n");
    }
  }
  else
  {
    for(i = 0; i < order.size(); ++i)
    {
      if(order.size() > 1)
        PrintConfigName(order[i]->F);
      order[i]->Compile(Root);
      if(order[i]->Comp)
        errors += order[i]->Comp->Errors;
      else
        missing_comp.push_back(order[i]->F);

      if(i + 1 < order.size())
        printf("\n\n");
    }

    if(errors.size())
    {
      printf("\n\n");
      PrintConfigName("Failed to Compile:");
      for(i = 0; i < errors.size(); ++i)
        printf("  %s\n", errors[i].c_str());
    }

    if(missing_comp.size())
    {
      printf("\n\n");
      PrintConfigName("No compiler was specified for:");
      for(i = 0; i < missing_comp.size(); ++i)
        printf("  %s\n", missing_comp[i].c_str());
    }
  }
}

bool DirectoryTree::IsCompiler(const CString &comp) const
{
  unsigned int i;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Configs.size(); ++i)
      VALIDATE_POINTER(Configs[i])}
  #endif

  for(i = 0; i < Configs.size(); ++i)
    if(Configs[i]->IsCompiler(comp))
      return 1;
  return 0;
}

void DirectoryTree::Clean(void)
{
  unsigned int i, j;
  CVector<CString> activelist = GetActiveCompilers();
  CString active;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Configs.size(); ++i)
      VALIDATE_POINTER(Configs[i])}
  #endif

  if(activelist.size())
    active = activelist[0];

  CVector<CString> comps = GetCompilerList();
  for(i = 0; i < comps.size(); ++i)
  {
    for(j = 0; j < Configs.size(); ++j)
      Configs[j]->SetCompiler(comps[i]);

    Root->Clean();

    for(j = 0; j < Configs.size(); ++j)
      Configs[j]->DBase.Clean();
  }

  for(i = 0; i < Configs.size(); ++i)
    Configs[i]->SetCompiler(active);
}

void DirectoryTree::SetCompiler(const CString &comp)
{
  unsigned int i, j;
  CVector<CString> active_comp = GetActiveCompilers();

  if(active_comp.size() == 1 && Cmp(active_comp[0].c_str(), comp.c_str()))
    return;

  VALIDATE_POINTER(Root)
  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Configs.size(); ++i)
      VALIDATE_POINTER(Configs[i])}
  #endif

  CVector<CString> comps = GetCompilerList();
  for(i = 0; i < comps.size(); ++i)
  {
    for(j = 0; j < Configs.size(); ++j)
      Configs[j]->SetCompiler(comps[i]);

    Root->Clean();

      // Don't clean the databases.
  }

  for(i = 0; i < Configs.size(); ++i)
    Configs[i]->SetCompiler(comp);

  BuildLibLists();
}

CVector<CString> DirectoryTree::GetCompilerList(void)
{
  unsigned int i, j, k;
  CVector<CString> ret;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Configs.size(); ++i)
      VALIDATE_POINTER(Configs[i])}
  #endif

  for(i = 0; i < Configs.size(); ++i)
    for(j = 0; j < Configs[i]->Comps.size(); ++j)
    {
      CString *name = &Configs[i]->Comps[j].Name;
      for(k = 0; k < ret.size(); ++k)
        if(Cmp(ret[k].c_str(), name->c_str()))
          break;

      if(k >= ret.size())
        ret.push_back(*name);
    }

  return ret;
}

CVector<CString> DirectoryTree::GetActiveCompilers(void)
{
  unsigned int i, j;
  CVector<CString> ret;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Configs.size(); ++i)
      VALIDATE_POINTER(Configs[i])}
  #endif

  for(i = 0; i < Configs.size(); ++i)
  {
    CString name;

    if(Configs[i]->Comp)
      name = Configs[i]->Comp->Name;
    else
      name = "no compiler";

    for(j = 0; j < ret.size(); ++j)
      if(Cmp(ret[j].c_str(), name.c_str()))
        break;

    if(j >= ret.size())
      ret.push_back(name);
  }

  return ret;
}

CVector<CString> DirectoryTree::GetSourceList(void)
{
  CVector<CString> ret;

  VALIDATE_POINTER(Root)

  Root->GetSourceList(ret);
  return ret;
}

CVector<CString> DirectoryTree::GetModifiedList(void)
{
  CVector<CString> ret;

  VALIDATE_POINTER(Root)

  Root->GetModifiedList(ret);
  return ret;
}

CVector<CString> DirectoryTree::GetDBBSList(void)
{
  CVector<CString> ret;
  unsigned int i;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {for(unsigned int i = 0; i < Configs.size(); ++i)
      VALIDATE_POINTER(Configs[i])}
  #endif

  for(i = 0; i < Configs.size(); ++i)
    ret.push_back(Configs[i]->F);

  return ret;
}

CVector<CString> DirectoryTree::GetOutputList(void)
{
  CVector<CString> ret;
  unsigned int i;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {
      for(unsigned int i = 0; i < Configs.size(); ++i)
      {
        VALIDATE_POINTER(Configs[i])
        if(Configs[i]->Comp)
          VALIDATE_POINTER(Configs[i]->Comp)
      }
    }
  #endif

  for(i = 0; i < Configs.size(); ++i)
    if(Configs[i]->Comp)
      if(Configs[i]->IsLibrary)
        ret.push_back(Configs[i]->LibDir + Configs[i]->Comp->FileTitle);
      else
        ret.push_back(Configs[i]->BinDir + Configs[i]->Comp->FileTitle);

  return ret;
}

void DirectoryTree::PrintDependencies(void)
{
  unsigned int i, j;

  #if CMM_ENABLED | SAFEALLOC_ENABLED
    {
      for(unsigned int i = 0; i < Configs.size(); ++i)
      {
        VALIDATE_POINTER(Configs[i])
        if(Configs[i]->Comp)
          VALIDATE_POINTER(Configs[i]->Comp)
      }
    }
  #endif

  printf("Dependencies:\n");
  for(i = 0; i < Configs.size(); ++i)
  {
    if(Configs[i]->Comp)
      if(Configs[i]->IsLibrary)
        printf("  %s\n", (Configs[i]->LibDir + Configs[i]->Comp->FileTitle).c_str());
      else
        printf("  %s\n", (Configs[i]->BinDir + Configs[i]->Comp->FileTitle).c_str());
    else
      printf("  %s\n", Configs[i]->F.c_str());
    for(j = 0; j < Configs[i]->AdditionalLibs.size(); ++j)
      printf("    %s\n", Configs[i]->AdditionalLibs[j].c_str());
    printf("\n");
  }

  if(!i)
    printf("\n");
}

void PrintConfigName(const CString &file)
{
  unsigned int len = file.length();
  CString padding(len + 4, '-');

  printf("%s\n", padding.c_str());
  printf("- %s -\n", file.c_str());
  printf("%s\n", padding.c_str());
}
