#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "Containers/CString.h"
#include "Directory/DirectoryTree.h"
#include "FileIO/Parser.h"           // Cmp()

#include "ConfigFile.h"

#include "Debug/Debug.h"

#define VER_MAJOR 0
#define VER_MINOR 8

int  ProcessSwitches(int argc, char *argv[], DirectoryTree &tree);
void PrintHelp(char *argv[]);
int  PrintStringList(const char *title, const CVector<CString> &str);

int main(int argc, char *argv[])
{
  MM_START()
  {
    DirectoryTree tree;

    srand((unsigned int)time(0));

    tree.BuildTree("");

    if(tree.Configs.size() && tree.Configs[0]->Comp)
    {
      VALIDATE_POINTER(tree.Configs[0]);
      VALIDATE_POINTER(tree.Configs[0]->Comp);
      tree.SetCompiler(tree.Configs[0]->Comp->Name);
    }

    switch(ProcessSwitches(argc, argv, tree))
    {
      case 0:
        tree.SaveConfigs();
        break;

      case 1:
        tree.Compile();
        tree.SaveConfigs();
        break;

      case 2:
        break;
    }
  }
  
  #if SAFEALLOC_ENABLED
    CheckAllPadding();
    DumpAllocatedMemMin();
    DumpAllocStatus();
  #endif

  return 0;
}

int ProcessSwitches(int argc, char *argv[], DirectoryTree &tree)
{
  int comp = 1;
  int i;
  for(i = 1; i < argc; ++i)
  {
    CString s(argv[i]);
    if(tree.IsCompiler(s))
    {
      tree.SetCompiler(s);
      return 1;
    }
    else
    if(Cmp(argv[i], "Clean"))
    {
      printf("Cleaning...");
      fflush(stdout);
      tree.Clean();
      printf(" Done\n");
      return 0;
    }
    else
    if(Cmp(argv[i], "Active"))
      comp = PrintStringList("Active Compiler(s)", tree.GetActiveCompilers());
    else
    if(Cmp(argv[i], "Compilers"))
      comp = PrintStringList("Compilers", tree.GetCompilerList());
    else
    if(Cmp(argv[i], "Files") || Cmp(argv[i], "Sources"))
      comp = PrintStringList("Source Files", tree.GetSourceList());
    else
    if(Cmp(argv[i], "Status") || Cmp(argv[i], "Modified"))
      comp = PrintStringList("Modified Files", tree.GetModifiedList());
    else
    if(Cmp(argv[i], "DBBS"))
      comp = PrintStringList(".dbbs Configuration Files", tree.GetDBBSList());
    else
    if(Cmp(argv[i], "Out"))
      comp = PrintStringList("Output binary files", tree.GetOutputList());
    else
    if(Cmp(argv[i], "Depends"))
    {
      tree.PrintDependencies();
      comp = 2;
    }
    else
    if(Cmp(argv[i], "Help")   || Cmp(argv[i], "-?")  || Cmp(argv[i], "/?") ||
       Cmp(argv[i], "--Help") || Cmp(argv[i], "--?") || Cmp(argv[i], "\\?"))
    {
      PrintHelp(argv);
      return 2;
    }
    else
    {
      PrintHelp(argv);
      printf("Error: Unknown Switch: %s\n", s.c_str());
      return 2;
    }
  }

  return comp;
}

void PrintHelp(char *argv[])
{
  printf("Directory Based Build System %d.%d\n", VER_MAJOR, VER_MINOR);
  printf("By: Craig Williams\n");
  printf("\n");
  printf("Usage:\n");
  printf("  %s <compiler> <commands(s)>\n", argv[0]);
  printf("\n");
  printf("<Command>: <Description>\n");
  printf("  %9s: %s\n", "Clean",     "Removes all object files and cleans all datatbases");
  printf("  %9s: %s\n", "Active",    "Lists the active compiler(s)");
  printf("  %9s: %s\n", "Compilers", "Lists all found compilers");
  printf("  %9s: %s\n", "Files",     "Lists all found source files");
  printf("  %9s: %s\n", "Sources",   "  Same as Files");
  printf("  %9s: %s\n", "Status",    "Lists all modified files");
  printf("  %9s: %s\n", "Modified",  "  Same as status");
  printf("  %9s: %s\n", "DBBS",      "Lists all found .dbbs files");
  printf("  %9s: %s\n", "Out",       "Lists all generated executables and libraries");
  printf("  %9s: %s\n", "Depends",   "Lists all library dependices for each exe and lib");
  printf("  %9s: %s\n", "Help",      "Display this message");
  printf("\n");
  printf("If a compiler is not specified, a default dbbs file will be generated\n");
  printf("\n\n");
}

int PrintStringList(const char *title, const CVector<CString> &str)
{
  unsigned int i;
  printf("%s:\n", title);
  for(i = 0; i < str.size(); ++i)
    printf("  %s\n", str[i].c_str());
  printf("\n");

  return 2;
}
